from typing import Tuple

import numpy as np

from ..base import LinkFunction
from ..types import ParameterShapes


class MatrixDiag(LinkFunction):
    """
    Wraps a link functions to be applied only on the diagonal of a square matrix.
    """

    valid_shapes = [ParameterShapes.DIAGONAL_MATRIX]

    def __init__(self, diag_link: LinkFunction, other_val=0):
        self.diag_link = diag_link
        self.other_val = other_val

    @property
    def link_support(self) -> Tuple[float, float]:
        return (self.diag_link.link_support[0], self.diag_link.link_support[1])

    def _make_indices(self, x: np.ndarray) -> Tuple:
        d = x.shape[1]
        i = np.diag_indices(d)
        return i

    def link(self, x: np.ndarray) -> np.ndarray:
        i = self._make_indices(x)
        out = np.full_like(x, self.other_val)
        out[:, i[0], i[1]] = self.diag_link.link(x[:, i[0], i[1]])
        return out

    def inverse(self, x: np.ndarray) -> np.ndarray:
        i = self._make_indices(x)
        out = np.full_like(x, self.other_val)
        out[:, i[0], i[1]] = self.diag_link.inverse(x[:, i[0], i[1]])
        return out

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        i = self._make_indices(x)
        out = np.full_like(x, self.other_val)
        out[:, i[0], i[1]] = self.diag_link.link_derivative(x[:, i[0], i[1]])
        return out

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        i = self._make_indices(x)
        out = np.full_like(x, self.other_val)
        out[:, i[0], i[1]] = self.diag_link.link_second_derivative(x[:, i[0], i[1]])
        return out

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        i = self._make_indices(x)
        out = np.full_like(x, self.other_val)
        out[:, i[0], i[1]] = self.diag_link.inverse_derivative(x[:, i[0], i[1]])
        return out

    def element_link(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link(x)
        else:
            raise ValueError("Element does not exist in the diagonal.")

    def element_link_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link_derivative(x)
        else:
            raise ValueError("Element does not exist in the diagonal.")

    def element_link_second_derivative(
        self, x: np.ndarray, i: int, j: int
    ) -> np.ndarray:
        if i == j:
            return self.diag_link.link_second_derivative(x)
        else:
            raise ValueError("Element does not exist in the diagonal.")

    def element_inverse_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.inverse_derivative(x)
        else:
            raise ValueError("Element does not exist in the diagonal.")

    def element_inverse(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.inverse(x)
        else:
            raise ValueError("Element does not exist in the diagonal.")


class MatrixDiagTriu(LinkFunction):
    """
    Wraps two link functions to be applied on the diagonal and the upper diagonal of a square matrix.
    """

    valid_shapes = [ParameterShapes.UPPER_TRIANGULAR_MATRIX]

    def __init__(self, diag_link: LinkFunction, triu_link: LinkFunction):
        self.diag_link = diag_link
        self.triu_link = triu_link

    @property
    def link_support(self) -> Tuple[float, float]:
        return (
            min(self.triu_link.link_support[0], self.diag_link.link_support[0]),
            max(self.triu_link.link_support[1], self.diag_link.link_support[1]),
        )

    def _make_indices(self, x: np.ndarray) -> Tuple:
        d = x.shape[1]
        i = np.diag_indices(d)
        j = np.triu_indices(d, k=1)
        return i, j

    def link(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.triu_link.link(x[:, j[0], j[1]])
        return out

    def inverse(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.inverse(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.triu_link.inverse(x[:, j[0], j[1]])
        return out

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.triu_link.link_derivative(x[:, j[0], j[1]])
        return out

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link_second_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.triu_link.link_second_derivative(x[:, j[0], j[1]])
        return out

    def inverse_derivative(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.inverse_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.triu_link.inverse_derivative(x[:, j[0], j[1]])
        return out

    def element_link(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link(x)
        else:
            return self.triu_link.link(x)

    def element_link_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link_derivative(x)
        else:
            return self.triu_link.link_derivative(x)

    def element_link_second_derivative(
        self, x: np.ndarray, i: int, j: int
    ) -> np.ndarray:
        if i == j:
            return self.diag_link.link_second_derivative(x)
        else:
            return self.triu_link.link_second_derivative(x)

    def element_inverse_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.element_inverse_derivative(x)
        else:
            return self.triu_link.element_inverse_derivative(x)

    def element_inverse(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.inverse(x)
        else:
            return self.triu_link.inverse(x)


class MatrixDiagTril(LinkFunction):
    valid_shapes = [ParameterShapes.LOWER_TRIANGULAR_MATRIX]

    def __init__(self, diag_link: LinkFunction, tril_link: LinkFunction):
        self.diag_link = diag_link
        self.tril_link = tril_link

    @property
    def link_support(self) -> Tuple[float, float]:
        return (
            min(self.tril_link.link_support[0], self.diag_link.link_support[0]),
            max(self.tril_link.link_support[1], self.diag_link.link_support[1]),
        )

    def _make_indices(self, x: np.ndarray) -> Tuple:
        d = x.shape[1]
        i = np.diag_indices(d)
        j = np.tril_indices(d, k=-1)
        return i, j

    def link(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.tril_link.link(x[:, j[0], j[1]])
        return out

    def inverse(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.inverse(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.tril_link.inverse(x[:, j[0], j[1]])
        return out

    def link_derivative(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.tril_link.link_derivative(x[:, j[0], j[1]])
        return out

    def inverse_derivative(self, x):
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.inverse_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.tril_link.inverse_derivative(x[:, j[0], j[1]])
        return out

    def link_second_derivative(self, x: np.ndarray) -> np.ndarray:
        i, j = self._make_indices(x)
        out = np.zeros_like(x)
        out[:, i[0], i[1]] = self.diag_link.link_second_derivative(x[:, i[0], i[1]])
        out[:, j[0], j[1]] = self.tril_link.link_second_derivative(x[:, j[0], j[1]])
        return out

    def element_link(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link(x)
        else:
            return self.tril_link.link(x)

    def element_link_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.link_derivative(x)
        else:
            return self.tril_link.link_derivative(x)

    def element_link_second_derivative(
        self, x: np.ndarray, i: int, j: int
    ) -> np.ndarray:
        if i == j:
            return self.diag_link.link_second_derivative(x)
        else:
            return self.tril_link.link_second_derivative(x)

    def element_inverse_derivative(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.inverse_derivative(x)
        else:
            return self.tril_link.inverse_derivative(x)

    def element_inverse(self, x: np.ndarray, i: int, j: int) -> np.ndarray:
        if i == j:
            return self.diag_link.inverse(x)
        else:
            return self.tril_link.inverse(x)



from typing import Tuple

import numpy as np
import scipy.special as spc
import scipy.stats as st

from ..base import Distribution, LinkFunction
from ..links import Log, Logit
from ..types import ParameterShapes


class ZeroAdjustedGamma(Distribution):
    """The Zero Adjusted Gamma Distribution for GAMLSS.
    
    The zero adjusted gamma distribution is a mixture of a discrete value 0 with
    probability \\nu, and a gamma GA(\\mu; \\sigma) distribution on the positive real line (0, \\infty)
    with probability (1 - \\nu).
    
    $$
    f_Y(y \\mid \\mu, \\sigma, \\nu) = 
    \\begin{cases}
        \\nu & \\text{if } y = 0 \\
        (1 - \\nu) f_W(y \mid \mu, \sigma) & \\text{if } y > 0
    \end{cases} 
    $$
    
    where $y$ is the observed data, $\\mu > 0$ is the location parameter,
    $\\sigma > 0$ is the scale parameter,
    and $\\nu \\in [0, \\infty) $  is the inflation parameter.

    """

    corresponding_gamlss: str = "ZAGA"
    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
        2: (np.nextafter(0, 1), np.nextafter(1, 0)),
    }
    distribution_support = (0, np.inf)
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
        2: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Log(),
        scale_link: LinkFunction = Log(),
        inflation_link: LinkFunction = Logit(),
    ) -> None:
        super().__init__(links={0: loc_link, 1: scale_link, 2: inflation_link})

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)

        if param == 0:
            result = (y - mu) / ((sigma**2) * (mu**2))

            return np.where(y == 0, 0, result)

        if param == 1:
            result = (2 / sigma**3) * (
                (y / mu)
                - np.log(y)
                + np.log(mu)
                + np.log(sigma**2)
                - 1
                + spc.digamma(1 / (sigma**2))
            )

            return np.where(y == 0, 0, result)

        if param == 2:
            return np.where(y == 0, 1 / nu, -1 / (1 - nu))

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma, nu = self.theta_to_params(theta)
        if param == 0:
            result = -1 / ((sigma**2) * (mu**2))

            return np.where(y == 0, 0, result)

        if param == 1:
            result = (4 / sigma**4) - (4 / sigma**6) * spc.polygamma(1, (1 / sigma**2))

            return np.where(y == 0, 0, result)

        if param == 2:
            return -1 / (nu * (1 - nu))

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        mu, sigma, nu = self.theta_to_params(theta)

        if sorted(params) == [0, 1]:
            return np.zeros_like(y)

        if sorted(params) == [0, 2]:
            return np.zeros_like(y)

        if sorted(params) == [1, 2]:
            return np.zeros_like(y)

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        return np.tile([np.mean(y), 1, 0.5], (y.shape[0], 1))

    def cdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2
        cdf_gamma = st.gamma(a=shape, loc=0, scale=scale).cdf(y)
        cont_cdf = nu + ((1 - nu) * cdf_gamma)

        result = np.where(y == 0, nu, cont_cdf)

        return result

    def pdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)

        result = (
            (1 - nu)
            * (y / (mu * sigma**2)) ** (1 / sigma**2)
            * (np.exp(-y / (mu * sigma**2)))
            / (y * spc.gamma(1 / sigma**2))
        )

        return np.where(y == 0, nu, result)

    def ppf(self, q, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2

        result = st.gamma(a=shape, loc=0, scale=scale).ppf((q - nu) / (1 - nu))

        return np.where(q <= nu, 0, result)

    def pmf(self, y, theta):
        raise NotImplementedError("PMF is not implemented for mixed distributions.")

    def rvs(self, size, theta):
        sim_unif = st.uniform.rvs(size=size)
        return self.ppf(q=sim_unif, theta=theta)

    def logpdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)

        result = (
            np.log(1 - nu)
            + (1 / sigma**2) * np.log(y / (mu * sigma**2))
            - y / (mu * sigma**2)
            - np.log(y)
            - spc.gammaln(1 / sigma**2)
        )

        return np.where(y == 0, np.log(nu), result)

    def logcdf(self, y, theta):
        mu, sigma, nu = self.theta_to_params(theta)
        shape = 1 / sigma**2
        scale = mu * sigma**2
        cdf_gamma = st.gamma(a=shape, loc=0, scale=scale).cdf(y)
        cont_logcdf_gamma = np.log(nu + ((1 - nu) * cdf_gamma))

        result = np.where(y == 0, np.log(nu), cont_logcdf_gamma)

        return result

    def logpmf(self, y, theta):
        return super().logpmf(y, theta)

    def calculate_conditional_initial_values(self, y, theta, param):
        return super().calculate_conditional_initial_values(y, theta, param)



from .beta import Beta
from .betainflated import BetaInflated
from .betainflatedzero import BetaInflatedZero
from .exponential import Exponential
from .gamma import Gamma
from .gumbel import Gumbel
from .inversegamma import InverseGamma
from .inversegaussian import InverseGaussian
from .johnsonsu import JSU
from .logistic import Logistic
from .lognormal import LogNormal
from .lognormalmedian import LogNormalMedian
from .mv_normal_chol import MultivariateNormalInverseCholesky
from .mv_normal_low_rank import MultivariateNormalInverseLowRank
from .mv_normal_modchol import MultivariateNormalInverseModifiedCholesky
from .mv_t_chol import MultivariateStudentTInverseCholesky
from .mv_t_low_rank import MultivariateStudentTInverseLowRank
from .mv_t_modchol import MultivariateStudentTInverseModifiedCholesky
from .normal import Normal, NormalMeanVariance
from .powerexponential import PowerExponential
from .reversegumbel import ReverseGumbel
from .skew_t import SkewT, SkewTMeanStd
from .studentt import StudentT
from .weibull import Weibull
from .zeroadjustedgamma import ZeroAdjustedGamma

__all__ = [
    "Normal",
    "NormalMeanVariance",
    "StudentT",
    "SkewT",
    "SkewTMeanStd",
    "JSU",
    "BetaInflated",
    "Gamma",
    "Beta",
    "LogNormal",
    "LogNormalMedian",
    "Logistic",
    "Exponential",
    "PowerExponential",
    "Gumbel",
    "InverseGaussian",
    "ReverseGumbel",
    "InverseGamma",
    "MultivariateNormalInverseCholesky",
    "MultivariateNormalInverseLowRank",
    "MultivariateNormalInverseModifiedCholesky",
    "MultivariateStudentTInverseCholesky",
    "MultivariateStudentTInverseLowRank",
    "MultivariateStudentTInverseModifiedCholesky",
    "BetaInflatedZero",
    "ZeroAdjustedGamma",
    "Weibull",
]



from itertools import product
from typing import Dict

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, MatrixDiag
from ..types import ParameterShapes


class MultivariateNormalInverseLowRank(MultivariateDistributionMixin, Distribution):
    """
    The multivariate normal (Gaussian) distribution parameterized by a low-rank precision matrix.

    The PDF of the multivariate normal distribution is given by:
    $$
    p(y \\mid \\mu, D, V) =
    |D + V V^T|^{1/2} \cdot (2\\pi)^{-k/2}
    \\exp\\left(-\\frac{1}{2} (y - \\mu)^T (D + V V^T) (y - \\mu)\\right)
    $$

    where \( k \) is the dimensionality of the data, \( \\mu \) is the location parameter, \( D \) is a diagonal matrix, and \( V \) is a low-rank matrix such that the precision is \( D + V V^T \).
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "diag_matrix", 2: "low_rank_matrix"}
    parameter_support = {0: (-np.inf, np.inf), 1: (0, np.inf), 2: (-np.inf, np.inf)}
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_normal
    scipy_names = {"mu": "mean", "sigma": "cov"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.DIAGONAL_MATRIX,
        2: ParameterShapes.MATRIX,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link_1: LinkFunction = MatrixDiag(Log()),
        scale_link_2: LinkFunction = Identity(),
        rank: int = 3,
    ):
        super().__init__(
            links={
                0: loc_link,
                1: scale_link_1,
                2: scale_link_2,
            }
        )
        self.rank = rank
        self._regularization_allowed = {0: False, 1: False, 2: True}

    def fitted_elements(self, dim: int):
        return {0: dim, 1: dim, 2: dim * self.rank}

    @property
    def param_structure(self):
        return self._param_structure

    def get_adr_regularization_distance(self, dim: int, param: int) -> np.ndarray:
        if param in (0, 1):
            return None
        if param == 2:
            return np.concatenate([np.repeat(i + 1, dim) for i in range(self.rank)])

    def get_regularization_size(self, dim: int) -> int:
        return self.rank + 1

    def index_flat_to_cube(self, k: int, d: int, param: int):
        if param == 0:
            return k
        if param == 1:
            i, j = np.diag_indices(d)
            return i[k], j[k]
        if param == 2:
            idx = [(j, i) for i, j in product(range(self.rank), range(d))]
            return idx[k][0], idx[k][1]

    def set_theta_element(
        self, theta: Dict, value: np.ndarray, param: int, k: int
    ) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if param == 0:
            theta[param][:, k] = value
        if (param == 1) | (param == 2):
            d = theta[0].shape[1]
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            theta[param][:, i, j] = value
        return theta

    def theta_to_params(self, theta: Dict[int, np.ndarray]):
        loc = theta[0]
        mat_d = theta[1]
        mat_v = theta[2]
        return loc, mat_d, mat_v

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]) -> Dict[str, np.ndarray]:
        """Map theta to `scipy` distribution parameters for the multivariate normal distribution.

        Args:
            theta (Dict[int, np.ndarray]): Fitted / predicted theta.

        Returns:
            Dict[str, np.ndarray]: Mapped predicted
        """
        out = {
            "mean": theta[0],
            "cov": np.linalg.inv(theta[1] + theta[2] @ theta[2].swapaxes(-1, -2)),
        }
        return out

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dp2(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = False
    ):
        mu, mat_d, mat_v = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = _deriv1_mu(y, mu, mat_d, mat_v, i=k)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = _deriv1_dmat(y, mu, mat_d, mat_v, i=i)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = _deriv1_vmat(y, mu, mat_d, mat_v, i=i, j=j)
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = True
    ):
        mu, mat_d, mat_v = self.theta_to_params(theta)
        d = y.shape[1]
        if param == 0:
            deriv = _deriv2_mu(y, mu, mat_d, mat_v, i=k)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = _deriv2_dmat(y, mu, mat_d, mat_v, i=i)
        if param == 2:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            deriv = _deriv2_vmat(y, mu, mat_d, mat_v, i=i, j=j)
        if clip:
            deriv = np.clip(deriv, -np.inf, -1e-5)

        return deriv

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link(y, i=i, j=j)

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link_derivative(y, i=i, j=j)

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link_second_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_link_second_derivative(y, i=i, j=j)

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].inverse(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_inverse(y, i=i, j=j)

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].inverse_derivative(y)
        if param == 1:
            i, j = self.index_flat_to_cube(k=k, d=d, param=param)
            return self.links[param].element_inverse_derivative(y, i=i, j=j)

    def initial_values(self, y: np.ndarray, param: int = 0):
        M = y.shape[0]
        if param == 0:
            return np.tile(np.mean(y, axis=0), (M, 1))
        if param == 1:
            mat_d = np.diag(1 / np.var(y, 0))
            return np.tile(mat_d, (M, 1, 1))
        if param == 2:
            M = y.shape[0]
            omega = np.linalg.inv(np.cov(y, rowvar=False))
            mat_d = np.diag(1 / np.var(y, 0))
            eig = np.linalg.eig(omega - mat_d)
            largest_ev = np.argsort(eig.eigenvalues)[-self.rank :][::-1]
            mat_v = eig.eigenvectors[:, largest_ev]
            return np.tile(mat_v, (M, 1, 1))

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param in (0, 2):
            return theta
        if param == 1:
            residuals = y - theta[0]
            mat_d = np.diag(1 / np.var(residuals, 0))
            theta[1] = np.tile(mat_d, (y.shape[0], 1, 1))
            return theta

    def cube_to_flat(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            return np.copy(x.diagonal(axis1=1, axis2=2))
        if param == 2:
            return x.swapaxes(-1, -2).reshape((x.shape[0], np.prod(x.shape[1:])))

    def flat_to_cube(self, x: np.ndarray, param: int):
        if param == 0:
            return x
        if param == 1:
            d = x.shape[1]
            out = np.zeros((x.shape[0], d, d))
            out[:, np.arange(d), np.arange(d)] = x
            return out
        if param == 2:
            d = int(x.shape[1] // self.rank)
            return x.reshape((x.shape[0], self.rank, d)).transpose(0, 2, 1)

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        loc, mat_d, mat_v = self.theta_to_params(theta)
        return _loglikelihood_gaussian_lowrank(y, loc, mat_d=mat_d, mat_v=mat_v)

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


def _loglikelihood_gaussian_lowrank(y, mu, mat_d, mat_v):
    """Fast evaluation of the batched log likelihood."""
    # k = y.shape[1]
    # cov = np.linalg.inv(mat_d + mat_v @ np.swapaxes(mat_v, -2, -1))
    # part1 = - k/2 * np.log(2 * np.pi)
    # part2 = - 1/2 * np.log(np.linalg.det(cov))
    # part3 = - 1 / 2 * np.sum((y - mu) * (np.linalg.inv(cov) @ (y - mu)[..., None]).squeeze(), 1)
    # return part1 + part2 + part3
    k = y.shape[1]
    precision = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    part1 = -k / 2 * np.log(2 * np.pi)
    part2 = 1 / 2 * np.log(np.linalg.det(precision))
    part3 = -1 / 2 * np.sum((y - mu) * (precision @ (y - mu)[..., None]).squeeze(), 1)
    return part1 + part2 + part3


def _deriv1_mu(y, mat_mu, mat_d, mat_v, i):
    term1 = mat_d[:, i, i] * (y - mat_mu)[:, i]
    term2 = np.sum(
        mat_v[:, [i], :] * mat_v * np.expand_dims(y - mat_mu, -1), axis=(-2, -1)
    )
    # This is a slightly cleaner version of below
    # term2 = np.squeeze(mat_v[:, [i], :] @ mat_v.swapaxes(-1, -2) @ np.expand_dims((y - mat_mu), -1))
    return term1 + term2


def _deriv2_mu(y, mat_mu, mat_d, mat_v, i):
    # Diagonal elements of the inverse covariance matrix
    return -(mat_d[:, i, i] + np.sum(mat_v[:, i] ** 2, 1))


def _deriv1_dmat(y, mat_mu, mat_d, mat_v, i):
    omega = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    part_1 = 0.5 * np.linalg.inv(omega)[:, i, i]
    part_2 = -0.5 * (y[:, i] - mat_mu[:, i]) ** 2
    return part_1 + part_2


def _deriv2_dmat(y, mat_mu, mat_d, mat_v, i):
    omega = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    cov = np.linalg.inv(omega)
    return 0.5 * -(cov[:, i, i] ** 2)


def _deriv1_vmat(y, mat_mu, mat_d, mat_v, i, j):
    # TODO: Would be nice to calculate only the necessary rows
    # of OMEGA in the future maybe!

    # Derivation for part 2
    # zzT @ V
    # zzT[:, i, :] @ mat_v[:, :, j]
    # select the correct row of zzT before
    # sum(z * z[:, i], axis=-1)
    omega = mat_d + mat_v @ np.swapaxes(mat_v, -2, -1)
    part_1 = np.sum(np.linalg.inv(omega)[:, i, :] * mat_v[:, :, j], axis=1)
    part_2 = -np.sum(
        (y - mat_mu) * np.expand_dims((y[:, i] - mat_mu[:, i]), -1) * mat_v[:, :, j], -1
    )
    return part_1 + part_2


def _deriv2_vmat(y, mat_mu, mat_d, mat_v, i, j):
    d = mat_d.shape[1]
    omega = mat_d + mat_v @ mat_v.swapaxes(-1, -2)
    cov = np.linalg.inv(omega)
    sum1 = 0
    sum2 = 0
    for k, q in product(range(d), range(d)):
        sum1 += cov[:, i, i] * mat_v[:, q, j] * cov[:, q, k] * mat_v[:, k, j]
        sum2 += cov[:, i, q] * mat_v[:, q, j] * cov[:, i, k] * mat_v[:, k, j]
    term1 = cov[:, i, i] - sum1 - sum2
    term2 = -((y - mat_mu)[:, i] ** 2)
    return term1 + term2



from .online_gamlss import OnlineDistributionalRegression
from .online_lasso import OnlineLasso
from .online_linear_model import OnlineLinearModel
from .online_mvdistreg import MultivariateOnlineDistributionalRegressionPath

__all__ = [
    "OnlineDistributionalRegression",
    "MultivariateOnlineDistributionalRegressionPath",
    "OnlineLasso",
    "OnlineLinearModel",
]



# Author Simon Hirsch
# MIT Licence
from typing import Dict

import numpy as np
import scipy.special as sp
import scipy.stats as st

from ..base import Distribution, LinkFunction, MultivariateDistributionMixin
from ..links import Identity, Log, LogShiftTwo, MatrixDiagTril
from ..types import ParameterShapes


class MultivariateStudentTInverseCholesky(MultivariateDistributionMixin, Distribution):
    # Covariance Matrix SIGMA
    # Precision Matrix OMEGA
    # Cholesky Factor inv(chol(SIGMA))
    # Lower diagonal matrix
    """
    The multivariate \\( t \\)-distribution parameterized by the inverse Cholesky factor of the precision (inverse scale) matrix.

    The PDF of the multivariate \\( t \\)-distribution is given by:
    $$
    p(y \mid \\mu, L, \\nu) =
    \\frac{\\Gamma\\left(\\frac{\\nu + k}{2}\\right)}
         {\\Gamma\\left(\\frac{\\nu}{2}\\right) \\left(\\pi \\nu\\right)^{k/2}}
    \\cdot |L|
    \\left(1 + \\frac{1}{\\nu} (y - \\mu)^T (L L^T) (y - \\mu)\\right)^{-\\frac{\\nu + k}{2}}
    $$

    where \\( k \\) is the dimensionality of the data, \\( \\mu \\) is the location parameter, \\( L \\) is the inverse Cholesky factor of the precision matrix (so the precision is \\( L L^T \\)), and \\( \\nu \\) is the degrees of freedom.
    """

    corresponding_gamlss: str = None
    parameter_names = {0: "mu", 1: "sigma", 2: "nu"}
    parameter_support = {0: (-np.inf, np.inf), 1: (-np.inf, np.inf), 2: (0, np.inf)}
    distribution_support = (-np.inf, np.inf)
    scipy_dist = st.multivariate_t
    scipy_names = {"mu": "loc", "sigma": "shape", "nu": "dof"}
    parameter_shape = {
        0: ParameterShapes.VECTOR,
        1: ParameterShapes.LOWER_TRIANGULAR_MATRIX,
        2: ParameterShapes.SCALAR,
    }

    def __init__(
        self,
        loc_link: LinkFunction = Identity(),
        scale_link: LinkFunction = MatrixDiagTril(Log(), Identity()),
        tail_link: LinkFunction = LogShiftTwo(),
        dof_guesstimate: float = 10.0,
    ):
        super().__init__(
            links={
                0: loc_link,
                1: scale_link,
                2: tail_link,
            }
        )
        self.is_multivariate = True
        self.dof_guesstimate = dof_guesstimate
        self.dof_independence = 1e6
        self._regularization_allowed = {0: False, 1: True, 2: False}

    @staticmethod
    def fitted_elements(dim: int):
        return {0: dim, 1: int(dim * (dim + 1) // 2), 2: 1}

    @property
    def param_structure(self):
        return self._param_structure

    def get_adr_regularization_distance(self, dim: int, param: int) -> float:
        if param in (0, 2):
            return None
        if param == 1:
            j, i = np.triu_indices(dim, k=0)
            return np.abs(i - j)

    def get_regularization_size(self, dim: int) -> int:
        return dim

    @staticmethod
    def index_flat_to_cube(k: int, d: int, param: int):
        if (param == 0) | (param == 2):
            return k
        if param == 1:
            # tril_indicies is row-wise
            # "inverted" triu_indicies is column-wise
            i, j = np.triu_indices(d)
            return j[k], i[k]

    @staticmethod
    def set_theta_element(theta: Dict, value: np.ndarray, param: int, k: int) -> Dict:
        """Sets an element of theta for parameter param and place k.

        !!! Note
            This will mutate `theta`!

        Args:
            theta (Dict): Current fitted $\theta$
            value (np.ndarray): Value to set
            param (int): Distribution parameter
            k (int): Flat element index $k$

        Returns:
            Dict: Theta where element (param, k) is set to value.
        """
        if (param == 0) | (param == 2):
            theta[param][:, k] = value
        if param == 1:
            d = theta[0].shape[1]
            i, j = np.triu_indices(d)
            theta[param][:, j[k], i[k]] = value
        return theta

    def theta_to_params(self, theta):
        loc = theta[0]
        chol = theta[1]
        dof = theta[2].squeeze()
        return loc, chol, dof

    def theta_to_scipy(self, theta: Dict[int, np.ndarray]):
        out = {
            "loc": theta[0],
            "shape": np.linalg.inv(theta[1].swapaxes(-1, -2) @ theta[1]),
            "df": theta[2],
        }
        return out

    def dl1_dp1(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def dl2_dp2(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented")

    def element_score(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl1_dp1(y=y, theta=theta, param=param, k=k)

    def element_hessian(self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0):
        return self.element_dl2_dp2(y=y, theta=theta, param=param, k=k)

    def element_dl1_dp1(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = False
    ):
        mu, chol, dof = self.theta_to_params(theta)
        if param == 0:
            deriv = first_partial_mu(y, mu, chol, dof, i=k)
        if param == 1:
            ## Weird notation in the Muchinsky Paper
            ii, jj = np.triu_indices(y.shape[1])
            deriv = first_partial_cov(y, mu, chol, dof, i=ii[k], j=jj[k])
        if param == 2:
            deriv = first_partial_dof(y, mu, chol, dof)
        return deriv

    def element_dl2_dp2(
        self, y: np.ndarray, theta: Dict, param: int = 0, k: int = 0, clip: bool = True
    ):
        mu, chol, dof = self.theta_to_params(theta)
        if param == 0:
            deriv = second_partial_mu(y, mu, chol, dof, i=k)
        if param == 1:
            ii, jj = np.triu_indices(y.shape[1])
            ## Weird notation in the Muchinsky Paper
            deriv = second_partial_cov(y, mu, chol, dof, i=ii[k], j=jj[k])
        if param == 2:
            deriv = second_partial_dof(y, mu, chol, dof)

        if clip:
            deriv = np.clip(deriv, -np.inf, -1e-5)

        return deriv

    def dl2_dpp(self, y: np.ndarray, theta: Dict, param: int = 0):
        raise NotImplementedError("Not implemented.")

    def element_link_function(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_link(y, i=j[k], j=i[k])

    def element_link_function_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_link_derivative(y, i=j[k], j=i[k])

    def element_link_function_second_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].link_second_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_link_second_derivative(y, i=j[k], j=i[k])

    def element_link_inverse(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].inverse(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_inverse(y, i=j[k], j=i[k])

    def element_link_inverse_derivative(
        self, y: np.ndarray, param: int = 0, k: int = 0, d: int = 0
    ) -> np.ndarray:
        if (param == 0) | (param == 2):
            return self.links[param].inverse_derivative(y)
        if param == 1:
            i, j = np.triu_indices(d)
            return self.links[param].element_inverse_derivative(y, i=j[k], j=i[k])

    def initial_values(self, y, param=0):
        M = y.shape[0]
        if param == 0:
            return np.tile(np.mean(y, axis=0), (M, 1))
        if param == 1:
            var = np.var(y, axis=0)
            shape = np.diag(var * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            chol = np.linalg.inv(np.linalg.cholesky(shape))
            return np.tile(chol, (M, 1, 1))
        if param == 2:
            return np.full((y.shape[0], 1), self.dof_independence)

    def set_initial_guess(
        self,
        y: np.ndarray,
        theta: Dict[int, np.ndarray],
        param: int,
    ) -> Dict[int, np.ndarray]:
        if param == 1:
            M = y.shape[0]
            residuals = y - theta[0]
            var = np.var(residuals, axis=0)
            shape = np.diag(var * (self.dof_guesstimate - 2) / self.dof_guesstimate)
            chol = np.linalg.inv(np.linalg.cholesky(shape))
            theta[1] = np.tile(chol, (M, 1, 1))
        if param == 2:
            theta[2] = np.full_like(theta[2], self.dof_guesstimate)

        return theta

    def cube_to_flat(self, x: np.ndarray, param: int):
        if (param == 0) | (param == 2):
            return x
        if param == 1:
            d = x.shape[1]
            i = np.triu_indices(d)
            out = x[:, i[1], i[0]]
            return out

    def flat_to_cube(self, x: np.ndarray, param: int):
        if (param == 0) | (param == 2):
            return x
        if param == 1:
            n, k = x.shape
            # The following conversion holds for upper diagonal matrices
            # We INCLUDE the diagonal!!
            # (D + 1) * D // 2 = k
            # (D + 1) * D = 2 * k
            # D^2 + D = 2 * k
            # ... Wolfram Alpha ...
            # D = 0.5 * (sqrt(8k + 1) - 1)
            d = int(1 / 2 * (np.sqrt(8 * k + 1) - 1))
            i = np.triu_indices(d)
            out = np.zeros((n, d, d))
            out[:, i[1], i[0]] = x
            return out

    def cdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pdf(self, y, theta):
        return np.exp(self.logpdf(y, theta))

    def ppf(self, q, theta):
        raise NotImplementedError("Not implemented")

    def rvs(self, size, theta):
        raise NotImplementedError("Not implemented")

    def logcdf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def logpdf(self, y, theta):
        loc, chol, dof = self.theta_to_params(theta)
        return batched_log_likelihood(y, loc, chol, dof)

    def logpmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def pmf(self, y, theta):
        raise NotImplementedError("Not implemented")

    def calculate_conditional_initial_values(
        self, y: np.ndarray, theta: Dict[int, np.ndarray]
    ) -> Dict[int, np.ndarray]:
        raise NotImplementedError("Not implemented")


def batched_log_likelihood(y, mu, chol, dof):
    D = y.shape[1]
    A = np.squeeze(sp.gammaln((dof + D) / 2))
    B = np.squeeze(sp.gammaln(dof / 2))
    C = 1 / 2 * D * np.squeeze(np.log(np.pi * dof))
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    return (
        A
        - B
        - C
        + np.log(np.linalg.det(chol))
        - np.squeeze((dof + D) / 2) * np.log((1 + np.squeeze((1 / dof)) * Z))
    )


def first_partial_mu(y, mu, chol, dof, i):
    k = y.shape[1]
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    precision = chol.transpose(0, 2, 1) @ chol
    part1 = (k + dof) / (2 * (Z + dof))
    part2 = 2 * np.sum(precision[:, i, :] * (y - mu), axis=1)
    return part1 * part2


def second_partial_mu(y, mu, chol, dof, i):
    k = y.shape[1]
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    precision = chol.transpose(0, 2, 1) @ chol

    deriv_1 = 2 * np.sum(precision[:, i, :] * (y - mu), axis=1)
    deriv_2 = 2 * precision[:, i, i]

    part1 = k + dof
    part2 = (Z + dof) * deriv_2 - deriv_1**2
    part3 = 2 * (Z + dof) ** 2

    return -(part1 * part2) / part3


def first_partial_cov(y, mu, chol, dof, i, j):
    k = y.shape[1]
    m = j + 1
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    # Calculate parts
    part2 = (k + dof) / (2 * (Z + dof))
    part3 = 2 * (y - mu)[:, i] * np.sum((y - mu)[:, :m] * chol[:, j, :m], axis=1)
    if i == j:
        part1 = 1 / chol[:, i, i]
    else:
        part1 = 0

    return part1 - part2 * part3


def second_partial_cov(y, mu, chol, dof, i, j):
    k = y.shape[1]
    m = j + 1
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)

    deriv_lambda_1 = (
        2 * (y - mu)[:, i] * np.sum((y - mu)[:, :m] * chol[:, j, :m], axis=1)
    )
    deriv_lambda_2 = 2 * (y - mu)[:, i] ** 2

    part2 = k + dof
    part3 = (Z + dof) * deriv_lambda_2 - deriv_lambda_1**2
    part4 = 2 * (Z + dof) ** 2

    if i == j:
        part1 = -1 / chol[:, i, i] ** 2
    else:
        part1 = 0

    return part1 - (part2 * part3) / part4


def first_partial_dof(y, mu, chol, dof):
    k = y.shape[1]
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    part1 = -(-dof * sp.digamma((k + dof) / 2) + k + dof * sp.digamma(dof / 2)) / (
        2 * dof
    )
    part2 = (Z * (k + dof)) / (2 * dof * (dof + Z)) - 1 / 2 * np.log((dof + Z) / dof)
    return part1 + part2


def second_partial_dof(y, mu, chol, dof):
    k = y.shape[1]
    Z = np.sum(np.einsum("ndk, nk -> nd", chol, (y - mu)) ** 2, axis=1)
    part1 = (
        1
        / 4
        * (
            (2 * k) / (dof**2)
            + sp.polygamma(1, (dof + k) / 2)
            - sp.polygamma(1, dof / 2)
        )
    )
    part2 = (Z * (dof * Z - k * (2 * dof + Z))) / (2 * dof**2 * (dof + Z) ** 2)
    return part1 + part2



from unittest.mock import patch

import numpy as np
import pytest
from sklearn.utils._tags import get_tags
from sklearn.utils.estimator_checks import check_estimator

from ondil.estimators import (
    MultivariateOnlineDistributionalRegressionPath,
    OnlineDistributionalRegression,
    OnlineLasso,
    OnlineLinearModel,
)
from ondil.scaler import OnlineScaler

EXPECTED_FAILED_CHECKS = {
    "check_sample_weight_equivalence_on_dense_data": "Insufficient data points to validate this check in the original scikit-learn implementation.",
}

# https://scikit-learn.org/stable/modules/generated/sklearn.utils.estimator_checks.check_estimator.html#sklearn.utils.estimator_checks.check_estimator


@pytest.mark.parametrize("scale_inputs", [False, True])
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("method", ["ols", "lasso", "elasticnet"])
@pytest.mark.parametrize("ic", ["aic", "bic", "hqc", "max"])
def test_sklearn_compliance_linear_model(scale_inputs, fit_intercept, method, ic):
    estimator = OnlineLinearModel(
        fit_intercept=fit_intercept,
        scale_inputs=scale_inputs,
        method=method,
        ic=ic,
    )
    check_estimator(
        estimator, on_fail="raise", expected_failed_checks=EXPECTED_FAILED_CHECKS
    )


@pytest.mark.parametrize("scale_inputs", [False, True])
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("ic", ["aic", "bic", "hqc", "max"])
@pytest.mark.parametrize("lambda_n", [50, 100])
@pytest.mark.parametrize("lambda_eps", [1e-4, 1e-2])
def test_sklearn_compliance_lasso_model(
    scale_inputs: bool,
    fit_intercept: bool,
    ic: str,
    lambda_n: int,
    lambda_eps: float,
):
    estimator = OnlineLasso(
        fit_intercept=fit_intercept,
        scale_inputs=scale_inputs,
        ic=ic,
        lambda_n=lambda_n,
        lambda_eps=lambda_eps,
    )
    check_estimator(
        estimator, on_fail="raise", expected_failed_checks=EXPECTED_FAILED_CHECKS
    )


@pytest.mark.parametrize("scale_inputs", [False, True])
@pytest.mark.parametrize("method", ["ols", "lasso", "elasticnet"])
@pytest.mark.parametrize("ic", ["aic", "bic", "hqc", "max"])
def test_sklearn_compliance_online_gamlss(scale_inputs, method, ic):
    estimator = OnlineDistributionalRegression(
        scale_inputs=scale_inputs,
        method=method,
        ic=ic,
    )
    check_estimator(
        estimator, on_fail="raise", expected_failed_checks=EXPECTED_FAILED_CHECKS
    )


def test_sklearn_compliance_scaler():
    estimator = OnlineScaler()
    check_estimator(
        estimator,
        on_fail="raise",
    )


# Monkey patching _enforce_estimator_tags_y to handle multivariate y
# This is necessary because the original implementation does not handle
# multivariate outputs correctly, leading to issues in the estimator checks.
# since sklearn thinks that y is always 1-D and it is sufficient to
# have a (n, 1) shape for y, while ondil assumes that y is at least bivariate
# in the sense of having at least (n, 2) shape.

# https://github.com/scikit-learn/scikit-learn/blob/da08f3d99194565caaa2b6757a3816eef258cd70/sklearn/utils/estimator_checks.py
# this is a helpful git link to understand the monkey patching


def test_sklearn_compliance_multivariate():
    EXPECTED_FAILED_CHECKS_MV = {
        "check_estimators_dtypes": "Weird data generation does not support multivariate properly.",
        "check_fit2d_1feature": "Does not work with my monkey patched version of scikit-learn _enforce_estimator_tags_y.",
    }

    def _enforce_estimator_tags_y(estimator, y):
        # Estimators with a `requires_positive_y` tag only accept strictly positive
        # data
        tags = get_tags(estimator)
        if tags.target_tags.positive_only:
            # Create strictly positive y. The minimal increment above 0 is 1, as
            # y could be of integer dtype.
            y += 1 + abs(y.min())
        if (
            tags.classifier_tags is not None
            and not tags.classifier_tags.multi_class
            and y.size > 0
        ):
            y = np.where(y == y.min(), y, y.min() + 1)
        # Estimators in mono_output_task_error raise ValueError if y is of 1-D
        # Convert into a 2-D y for those estimators.
        if tags.target_tags.multi_output and not tags.target_tags.single_output:
            y = np.vstack((y, np.roll(y, 2))).transpose()
        return y

    with patch(
        "sklearn.utils.estimator_checks._enforce_estimator_tags_y",
        new=_enforce_estimator_tags_y,
    ):
        _ = check_estimator(
            MultivariateOnlineDistributionalRegressionPath(verbose=0),
            on_fail="raise",
            expected_failed_checks=EXPECTED_FAILED_CHECKS_MV,
        )



from typing import Tuple

import numpy as np
import scipy.stats as st

from ..base import Distribution, LinkFunction, ScipyMixin
from ..links import Log
from ..types import ParameterShapes


class Weibull(ScipyMixin, Distribution):
    """The Weibull distribution parameterized by scale (mu) and shape (sigma).

    The probability density function is defined as:
    $$
    f(y | \\mu, \\sigma) = \\frac{\\sigma}{\\mu} \\left(\\frac{y}{\\mu}\\right)^{\\sigma-1} \\exp\\left[-\\left(\\frac{y}{\\mu}\\right)^{\\sigma}\\right]
    $$

    where $y > 0$, $\\mu > 0$ is the scale parameter, and $\\sigma > 0$ is the shape parameter.

    !!! Note
        This distribution corresponds to the WEI() distribution in GAMLSS.

        The parameterization matches scipy.stats.weibull_min with:
        - scale = mu (GAMLSS mu)
        - c = sigma (GAMLSS sigma, the shape parameter)

    Args:
        scale_link (LinkFunction, optional): The link function for $\\mu$. Defaults to Log().
        shape_link (LinkFunction, optional): The link function for $\\sigma$. Defaults to Log().
    """

    corresponding_gamlss: str = "WEI"
    parameter_names = {0: "mu", 1: "sigma"}
    parameter_support = {
        0: (np.nextafter(0, 1), np.inf),
        1: (np.nextafter(0, 1), np.inf),
    }
    distribution_support = (np.nextafter(0, 1), np.inf)
    parameter_shape = {
        0: ParameterShapes.SCALAR,
        1: ParameterShapes.SCALAR,
    }
    # Scipy equivalent and parameter mapping ondil -> scipy
    scipy_dist = st.weibull_min
    scipy_names = {"mu": "scale", "sigma": "c"}

    def __init__(
        self,
        scale_link: LinkFunction = Log(),
        shape_link: LinkFunction = Log(),
    ) -> None:
        """Initialize the Weibull distribution.

        Args:
            scale_link (LinkFunction, optional): Link function for mu (scale). Defaults to Log().
            shape_link (LinkFunction, optional): Link function for sigma (shape). Defaults to Log().
        """
        super().__init__(
            links={
                0: scale_link,
                1: shape_link,
            }
        )

    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            # First derivative with respect to mu (scale parameter)
            # dldm = ((y/mu)^sigma - 1) * (sigma/mu)
            return ((y / mu) ** sigma - 1) * (sigma / mu)

        if param == 1:
            # First derivative with respect to sigma (shape parameter)
            # dldd = 1/sigma - log(y/mu) * ((y/mu)^sigma - 1)
            return (1 / sigma) - np.log(y / mu) * ((y / mu) ** sigma - 1)

    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int = 0) -> np.ndarray:
        self._validate_dln_dpn_inputs(y, theta, param)
        mu, sigma = self.theta_to_params(theta)

        if param == 0:
            # Second derivative with respect to mu (scale parameter)
            # d2ldm2 = -sigma^2 / mu^2
            return -(sigma**2) / (mu**2)

        if param == 1:
            # Second derivative with respect to sigma (shape parameter)
            # d2ldd2 = -1.82368 / sigma^2
            # This is an approximation based on the expected Fisher information
            # The exact value would be -trigamma(1) ≈ -1.82368
            return -1.82368 / (sigma**2)

    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int] = (0, 1)
    ) -> np.ndarray:
        self._validate_dl2_dpp_inputs(y, theta, params)
        if sorted(params) == [0, 1]:
            # Cross derivative with respect to mu and sigma
            # d2ldmdd = 0.422784 / mu
            # This is an approximation based on the expected Fisher information
            # The exact value would be (digamma(1) - 1) ≈ 0.422784
            mu, _ = self.theta_to_params(theta)
            return 0.422784 / mu

    def initial_values(self, y: np.ndarray) -> np.ndarray:
        # Following GAMLSS initialization:
        # log.Y.m <- log(y)
        # var.Y.v <- var(log(y))
        # sd.Y.s <- 1.283/sqrt(var.Y.v)
        # mu <- exp(log.Y.m + 0.5772/sd.Y.s)
        # sigma <- 1.283/sqrt(var(log(y)))
        log_y = np.log(y)
        log_y_mean = np.mean(log_y)
        log_y_var = np.var(log_y, ddof=1)

        sigma_init = 1.283 / np.sqrt(log_y_var) if log_y_var > 0 else 1.0
        mu_init = np.exp(log_y_mean + 0.5772 / sigma_init)

        initial_params = [mu_init, sigma_init]
        return np.tile(initial_params, (y.shape[0], 1))



import numpy as np


def ensure_atleast_bivariate(y: np.ndarray) -> None:
    """Validates that the response variable is a multivariate array with at least two dimensions.

    Args:
        y (np.ndarray): The response variable to validate.
    Raises:
        ValueError: If `y` does not have at least two dimensions and has less than two columns.
    """
    if y.shape[1] < 2:
        raise ValueError(
            "Multivariate response must have at least two dimensions. "
            f"Got {y.shape[1]} dimensions."
        )


def validate_response_support(estimator, y: np.ndarray) -> None:
    """Validates that the response values are within the distribution support of the estimator.

    Args:
        estimator: The estimator with a distribution that defines the support.
        y (np.ndarray): The response variable to validate.
    Raises:
        ValueError: If any response values are outside the distribution support.

    """
    if np.any(y < estimator.distribution.distribution_support[0]):
        raise ValueError(
            f"Response values must be greater than or equal to {estimator.distribution.distribution_support[0]}. "
            f"Got values less than this: {y[y < estimator.distribution.distribution_support[0]]}."
        )
    if np.any(y > estimator.distribution.distribution_support[1]):
        raise ValueError(
            f"Response values must be less than or equal to {estimator.distribution.distribution_support[1]}. "
            f"Got values greater than this: {y[y > estimator.distribution.distribution_support[1]]}."
        )



import copy
import numbers
import time
from typing import Any, Dict, List, Literal, Optional, Tuple
import warnings

import numba as nb
import numpy as np
from sklearn.base import BaseEstimator, MultiOutputMixin, RegressorMixin, _fit_context
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_is_fitted, validate_data

from ..base import Distribution, OndilEstimatorMixin
from ..design_matrix import make_intercept
from ..distributions import MultivariateNormalInverseCholesky
from ..gram import init_forget_vector
from ..information_criteria import InformationCriterion
from ..methods import get_estimation_method
from ..scaler import OnlineScaler
from ..types import ParameterShapes
from ..utils import calculate_effective_training_length
from ..validation import ensure_atleast_bivariate, validate_response_support

try:
    import pandas as pd

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False

try:
    import polars as pl

    HAS_POLARS = True
except ImportError:
    HAS_POLARS = False


def indices_along_diagonal(D: int) -> List:
    """This functions generates a list of indices that will go along
    the diagonal first and then go along the off-diagonals for a upper/lower
    triangular matrix.

    Args:
        D (int): Dimension of the matrix.

    Returns:
        List: List of indices
    """
    K = []
    for i in range(D):
        K.append(i)
        for j in range(D, i + 1, -1):
            K.append((K[-1] + j))
    return K


@nb.jit(
    [
        "float32[:, :](float32[:], float32[:], float32[:, :], boolean)",
        "float64[:, :](float64[:], float64[:], float64[:, :], boolean)",
    ],
    parallel=False,
)
def fast_vectorized_interpolate(x, xp, fp, ascending=True):
    dim = fp.shape[1]
    out = np.zeros((x.shape[0], dim), dtype=fp.dtype)
    if not ascending:
        x_asc = x[::-1]
        xp_asc = xp[::-1]

    for i in range(dim):
        if ascending:
            out[:, i] = np.interp(
                x=x,
                xp=xp,
                fp=fp[:, i],
            )
        else:
            out[:, i] = np.interp(
                x=x_asc,
                xp=xp_asc,
                fp=fp[:, i][::-1],
            )[::-1]
    return out


def make_model_array(X, eq, fit_intercept):
    n = X.shape[0]

    if isinstance(eq, str) and (eq == "intercept"):
        if not fit_intercept:
            raise ValueError(
                "fit_intercept[param] is false, but equation says intercept."
            )
        out = make_intercept(n_observations=n)
    else:
        if isinstance(eq, str) and (eq == "all"):
            if isinstance(X, np.ndarray):
                out = X
            if HAS_PANDAS and isinstance(X, pd.DataFrame):
                out = X.to_numpy()
            if HAS_POLARS and isinstance(X, pl.DataFrame):
                out = X.to_numpy()
        elif isinstance(eq, np.ndarray) or isinstance(eq, list):
            if isinstance(X, np.ndarray):
                out = X[:, eq]
            if HAS_PANDAS and isinstance(X, pd.DataFrame):
                out = X.loc[:, eq]
            if HAS_POLARS and isinstance(X, pl.DataFrame):
                out = X.select(eq).to_numpy()
        else:
            raise ValueError("Did not understand equation. Please check.")

        if fit_intercept:
            out = np.hstack((make_intercept(n), out))

    return out


class MultivariateOnlineDistributionalRegressionPath(
    OndilEstimatorMixin, RegressorMixin, MultiOutputMixin, BaseEstimator
):
    _parameter_constraints = {
        "distribution": [callable],
        "equation": [dict, type(None)],
        "forget": [Interval(numbers.Real, 0.0, 1.0, closed="left"), dict],
        "learning_rate": [Interval(numbers.Real, 0.0, 1.0, closed="left")],
        "fit_intercept": [bool, dict],
        "regularize_intercept": [bool, dict],
        "scale_inputs": [bool, np.ndarray],
        "verbose": [Interval(numbers.Integral, 0, None, closed="left")],
        "method": [StrOptions({"ols", "lasso"}), dict],
        "ic": [StrOptions({"ll", "aic", "bic", "hqc", "max"})],
        "iteration_along_diagonal": [bool],
        "approx_fast_model_selection": [bool],
        "max_regularisation_size": [
            Interval(numbers.Integral, 1, None, closed="left"),
            type(None),
        ],
        "early_stopping": [bool],
        "early_stopping_criteria": [StrOptions({"ll", "aic", "bic", "hqc", "max"})],
        "early_stopping_abs_tol": [Interval(numbers.Real, 0.0, None, closed="left")],
        "early_stopping_rel_tol": [Interval(numbers.Real, 0.0, None, closed="left")],
        "weight_delta": [numbers.Real, dict],
        "max_iterations_inner": [Interval(numbers.Integral, 1, None, closed="left")],
        "max_iterations_outer": [Interval(numbers.Integral, 1, None, closed="left")],
        "overshoot_correction": [dict, type(None)],
        "dampen_estimation": [
            bool,
            Interval(numbers.Integral, 0, None, closed="left"),
            dict,
        ],
        "debug": [bool],
        "rel_tol_inner": [Interval(numbers.Real, 0.0, None, closed="left")],
        "abs_tol_inner": [Interval(numbers.Real, 0.0, None, closed="left")],
        "rel_tol_outer": [Interval(numbers.Real, 0.0, None, closed="left")],
        "abs_tol_outer": [Interval(numbers.Real, 0.0, None, closed="left")],
    }

    def __init__(
        self,
        distribution: Distribution = MultivariateNormalInverseCholesky(),
        equation: Dict | None = None,
        forget: float | Dict = 0.0,
        learning_rate: float = 0.0,
        fit_intercept: bool = True,
        regularize_intercept: bool = False,
        scale_inputs: bool = True,
        verbose: int = 1,
        method: Literal["ols", "lasso"] | Dict[int, Literal["ols", "lasso"]] = "ols",
        ic: Literal["ll", "aic", "bic", "hqc", "max"] = "aic",
        iteration_along_diagonal: bool = False,
        approx_fast_model_selection: bool = True,
        max_regularisation_size: int | None = None,
        early_stopping: bool = True,
        early_stopping_criteria: Literal["ll", "aic", "bic", "hqc", "max"] = "aic",
        early_stopping_abs_tol: float = 0.001,
        early_stopping_rel_tol: float = 0.001,
        weight_delta: float | Dict[int, float] = 1.0,
        max_iterations_inner: int = 10,
        max_iterations_outer: int = 10,
        overshoot_correction: Optional[Dict[int, float]] = None,
        dampen_estimation: bool | int = False,
        debug: bool = False,
        rel_tol_inner: float = 1e-3,
        abs_tol_inner: float = 1e-3,
        rel_tol_outer: float = 1e-3,
        abs_tol_outer: float = 1e-3,
    ):
        """Initialize the Online Multivariate Distributional Regression Path Estimator.

        This estimator fits a multivariate distribution to the data using an online approach.
        It regularizes the scale matrix by estimating it sequentially from an "independence" configuration to a full covariance matrix,
        using the AD-R regularization scheme. Supports various regression methods, model selection criteria, early stopping, and input scaling.

        Args:
            distribution (Distribution, optional): The multivariate distribution to fit. Defaults to MultivariateNormalInverseCholesky().
            equation (Dict | None, optional): Dictionary specifying the regression equation. Defaults to None.
            forget (float | Dict, optional): Forgetting factor for online updates, can be a float or per-dimension dict. Defaults to 0.0.
            learning_rate (float, optional): Learning rate for online updates. Defaults to 0.0.
            fit_intercept (bool, optional): Whether to fit an intercept term. Defaults to True.
            regularize_intercept (bool, optional): Whether to regularize the intercept term. Defaults to False.
            scale_inputs (bool, optional): Whether to scale input features. Defaults to True.
            verbose (int, optional): Verbosity level for logging. Defaults to 1.
            method (Literal["ols", "lasso"] | Dict[int, Literal["ols", "lasso"]], optional): Regression method(s) to use, either "ols", "lasso", or per-dimension dict. Defaults to "ols".
            ic (Literal["ll", "aic", "bic", "hqc", "max"], optional): Information criterion for model selection. Defaults to "aic".
            iteration_along_diagonal (bool, optional): If True, iterates regularization along the diagonal of the scale matrix. Defaults to False.
            approx_fast_model_selection (bool, optional): If True, uses approximate fast model selection. Defaults to True.
            max_regularisation_size (int | None, optional): Maximum size for regularization path. Defaults to None.
            early_stopping (bool, optional): Enables early stopping during regularization. Defaults to True.
            early_stopping_criteria (Literal["ll", "aic", "bic", "hqc", "max"], optional): Criterion for early stopping. Defaults to "aic".
            early_stopping_abs_tol (float, optional): Absolute tolerance for early stopping. Defaults to 0.001.
            early_stopping_rel_tol (float, optional): Relative tolerance for early stopping. Defaults to 0.001.
            weight_delta (float | Dict[int, float], optional): Step size for weight updates, can be float or per-dimension dict. Defaults to 1.0.
            max_iterations_inner (int, optional): Maximum number of inner iterations. Defaults to 10.
            max_iterations_outer (int, optional): Maximum number of outer iterations. Defaults to 10.
            overshoot_correction (Optional[Dict[int, float]], optional): Correction factors for overshooting during updates. Defaults to None.
            dampen_estimation (bool | int, optional): If True or int, dampens estimation updates. Defaults to False.
            debug (bool, optional): Enables debug mode for additional logging. Defaults to False.
            rel_tol_inner (float, optional): Relative tolerance for convergence in inner loop. Defaults to 1e-3.
            abs_tol_inner (float, optional): Absolute tolerance for convergence in inner loop. Defaults to 1e-3.
            rel_tol_outer (float, optional): Relative tolerance for convergence in outer loop. Defaults to 1e-3.
            abs_tol_outer (float, optional): Absolute tolerance for convergence in outer loop. Defaults to 1e-3.
        """

        self.distribution = distribution
        self.forget = forget
        self.fit_intercept = fit_intercept
        self.regularize_intercept = regularize_intercept
        self.dampen_estimation = dampen_estimation
        self.weight_delta = weight_delta

        self.method = method
        self.learning_rate = learning_rate
        self.equation = equation
        self.iteration_along_diagonal = iteration_along_diagonal
        self.overshoot_correction = overshoot_correction

        # Early stopping
        self.max_regularisation_size = max_regularisation_size
        self.early_stopping = early_stopping
        self.early_stopping_criteria = early_stopping_criteria
        self.early_stopping_abs_tol = early_stopping_abs_tol
        self.early_stopping_rel_tol = early_stopping_rel_tol

        # Scaler
        self.scale_inputs = scale_inputs

        # For LASSO
        self.ic = ic
        self.approx_fast_model_selection = approx_fast_model_selection

        # Iterations and other internal stuff
        self.max_iterations_outer = max_iterations_outer
        self.max_iterations_inner = max_iterations_inner
        self.rel_tol_inner = rel_tol_inner
        self.abs_tol_inner = abs_tol_inner
        self.rel_tol_outer = rel_tol_outer
        self.abs_tol_outer = abs_tol_outer

        # For pretty printing.
        self.verbose = verbose
        self.debug = debug

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.target_tags.two_d_labels = True
        tags.target_tags.single_output = False
        tags.target_tags.multi_output = True
        tags.input_tags.sparse = False
        tags.regressor_tags.poor_score = True
        return tags

    def _process_parameter(self, attribute: Any, default: Any, name: str) -> None:
        if isinstance(attribute, dict):
            for p in range(self.distribution.n_params):
                if p not in attribute.keys():
                    warnings.warn(
                        f"[{self.__class__.__name__}] "
                        f"No value given for parameter {name} for distribution "
                        f"parameter {p}. Setting default value {default}.",
                        RuntimeWarning,
                        stacklevel=1,
                    )
                    if isinstance(default, dict):
                        attribute[p] = default[p]
                    else:
                        attribute[p] = default
        else:
            # No warning since we expect that floats/strings/ints are either the defaults
            # Or given on purpose for all params the ame
            attribute = {p: attribute for p in range(self.distribution.n_params)}

        return attribute

    def _prepare_estimator(self):
        self._scaler = OnlineScaler(
            forget=self.learning_rate,
            to_scale=self.scale_inputs,
        )
        # For simplicity
        self._forget = self._process_parameter(
            self.forget,
            default=0.0,
            name="forget",
        )
        self._fit_intercept = self._process_parameter(
            self.fit_intercept,
            default=True,
            name="fit_intercept",
        )
        self._regularize_intercept = self._process_parameter(
            self.regularize_intercept,
            default=False,
            name="regularize_intercept",
        )
        self._dampen_estimation = self._process_parameter(
            self.dampen_estimation,
            default=False,
            name="dampen_estimation",
        )
        self._weight_delta = self._process_parameter(
            self.weight_delta,
            default=1.0,
            name="weight_delta",
        )
        self._overshoot_correction = self._process_parameter(
            self.overshoot_correction,
            default=None,
            name="overshoot_correction",
        )

    def _make_iteration_indices(self, param: int):
        if (
            self.distribution.parameter_shape
            in [
                ParameterShapes.SQUARE_MATRIX,
                ParameterShapes.LOWER_TRIANGULAR_MATRIX,
                ParameterShapes.UPPER_TRIANGULAR_MATRIX,
            ]
        ) & self.iteration_along_diagonal:
            index = np.arange(indices_along_diagonal(self.dim_[param]))
        else:
            index = np.arange(self.n_dist_elements_[param])

        return index

    def _make_matrix_or_intercept(
        self, n_observations: int, x: np.ndarray, add_intercept: bool, param: int
    ):
        """
        Make the covariate matrix or an intercept array based on the input.

        Args:
            y (np.ndarray): Response variable.
            x (np.ndarray): Covariate matrix.
            add_intercept (bool): Flag indicating whether to add an intercept.
            param (int): Parameter index.

        Returns:
            np.ndarray: Matrix or intercept array.
        """
        if x is None:
            return make_intercept(n_observations=n_observations)
        elif add_intercept[param]:
            return self._add_intercept(x=x, param=param)
        else:
            return x

    # Different UV - MV
    def _make_initial_theta(self, y: np.ndarray):
        theta = {
            a: {
                p: self.distribution.initial_values(y, p)
                for p in range(self.distribution.n_params)
            }
            for a in range(self.adr_steps_)
        }
        # Handle AD-R Regularization
        for a in range(self.adr_steps_):
            for p in range(self.distribution.n_params):
                theta = self._handle_path_regularization(theta=theta, p=p, a=a)

        return theta

    def _handle_path_regularization(self, theta, p: int, a: int):
        if self.distribution._regularization_allowed[p]:
            mask = (
                self._adr_distance[p] >= self._adr_mapping_index_to_max_distance[p][a]
            )
            regularized = self.distribution.cube_to_flat(theta[a][p], p)
            regularized[:, mask] = regularized[:, mask] = 0
            theta[a][p] = self.distribution.flat_to_cube(regularized, p)
            return theta
        else:
            # If no regularization is allowed, we just return the theta
            return theta

    # Only MV
    def _is_element_adr_regularized(self, p: int, k: int, a: int):
        if not self.distribution._regularization_allowed[p]:
            return False
        else:
            return (
                self._adr_distance[p][k]
                >= self._adr_mapping_index_to_max_distance[p][a]
            )

    # Only MV
    # This should be using the distribution
    def _prepare_adr_regularization(self) -> None:
        self._adr_distance = {}
        self._adr_mapping_index_to_max_distance = {
            p: np.arange(1, self.adr_steps_ + 1)
            for p in range(self.distribution.n_params)
            if self.distribution._regularization_allowed[p]
        }
        for p in range(self.distribution.n_params):
            if self.distribution._regularization_allowed[p]:
                self._adr_distance[p] = (
                    self.distribution.get_adr_regularization_distance(
                        dim=self.dim_, param=p
                    )
                )

    # Different UV-MV
    def _get_number_of_covariates(self, X: np.ndarray):
        J = {}
        for p in range(self.distribution.n_params):
            J[p] = {}
            for k in range(self.n_dist_elements_[p]):
                if isinstance(self._equation[p][k], str):
                    if self._equation[p][k] == "all":
                        J[p][k] = X.shape[1] + int(self._fit_intercept[p])
                    if self._equation[p][k] == "intercept":
                        J[p][k] = 1
                elif isinstance(self._equation[p][k], np.ndarray) or isinstance(
                    self._equation[p][k], list
                ):
                    J[p][k] = len(self._equation[p][k]) + int(self._fit_intercept[p])
                else:
                    raise ValueError("Something unexpected happened")
        return J

    def _prepare_estimation_method(self, y):
        dim = y.shape[1]
        method = {
            p: {k: get_estimation_method(m) for k in range(K)}
            for p, (m, K) in enumerate(
                zip(
                    self._process_parameter(
                        self.method, default="ols", name="method"
                    ).values(),
                    self.distribution.fitted_elements(dim).values(),
                )
            )
        }
        return method

    def _check_path_based_estimation_methods(self):
        self._lambda_n = {}

        for p in range(self.distribution.n_params):
            lmbdas = []
            for k in range(self.n_dist_elements_[p]):
                if self._method[p][k]._path_based_method:
                    lmbdas.append(self._method[p][k]._path_length)
            if len(set(lmbdas)) > 1:
                raise ValueError(
                    f"Distribution parameter {p} has different path lengths for different elements. "
                    "This is not supported by the current implementation. "
                    "Please use a single path length for all elements of the distribution parameter."
                )
            if len(set(lmbdas)) == 1:
                self._lambda_n[p] = lmbdas[0]
            else:
                # If no path length is given, we assume that the method is not path-based
                # and we set the path length to 1.
                self._lambda_n[p] = 1

    def _get_next_adr_start_values(self, theta, p, a):
        mask = (
            self._adr_distance[p] >= self._adr_mapping_index_to_max_distance[p][a - 1]
        )
        prev_est = self.distribution.cube_to_flat(theta[a - 1][p], p)
        regularized = self.distribution.cube_to_flat(theta[a][p], p)
        regularized[:, ~mask] = prev_est[:, ~mask]
        start_value = self.distribution.flat_to_cube(regularized, p)
        return start_value

    # Different UV-MV
    def validate_equation(self, equation):
        if equation is None:
            warnings.warn(
                f"[{self.__class__.__name__}] "
                "Equation is not specified. "
                "Per default, will estimate the first distribution parameter by all covariates found in X. "
                "All other distribution parameters will be estimated by an intercept."
            )
            equation = {
                p: (
                    {k: "all" for k in range(self.n_dist_elements_[p])}
                    if p == 0
                    else {k: "intercept" for k in range(self.n_dist_elements_[p])}
                )
                for p in range(self.distribution.n_params)
            }
        else:
            for p in range(self.distribution.n_params):
                # Check that all distribution parameters are in the equation.
                # If not, add intercept.
                if p not in equation.keys():
                    print(
                        f"{self.__class__.__name__}",
                        f"Distribution parameter {p} is not in equation.",
                        "All elements of the parameter will be estimated by an intercept.",
                    )
                    equation[p] = {
                        k: "intercept" for k in range(self.n_dist_elements_[p])
                    }

                else:
                    for k in range(self.n_dist_elements_[p]):
                        if k not in equation[p].keys():
                            message = (
                                f"Distribution parameter {p}, element {k} is not in equation.",
                                "Element of the parameter will be estimated by an intercept.",
                            )
                            self._print_message(level=1, message=message)
                            equation[p][k] = "intercept"

                    if not (
                        isinstance(equation[p][k], np.ndarray)
                        or (equation[p][k] in ["all", "intercept"])
                    ):
                        if not (
                            isinstance(equation[p], list) and (HAS_PANDAS | HAS_POLARS)
                        ):
                            raise ValueError(
                                "The equation should contain of either: \n"
                                " - a numpy array of dtype int, \n"
                                " - a list of string column names \n"
                                " - or the strings 'all' or 'intercept' \n"
                                f"you have passed {equation[p]} for the distribution parameter {p}."
                            )

        return equation

    # Different UV - MV
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(
        self, X: np.ndarray, y: np.ndarray
    ) -> "MultivariateOnlineDistributionalRegressionPath":
        """Fit the estimator to the data.

        !!! Note
            The response $y$ must be a multivariate response
            variable, i.e., a 2D array where each column represents
            a different response variable and we expect at least a
            bivariate response. Hence, passing an `(n, 1)` array
            for y does not work!

        Args:
            X (np.ndarray): Covariate or feature matrix.
            y (np.ndarray): Multivariate response variable.

        Returns:
            estimator: Returns the fitted estimator.
        """

        X, y = validate_data(
            self,
            X=X,
            y=y,
            reset=True,
            dtype=[np.float64, np.float32],
            multi_output=True,
            ensure_min_samples=2,
        )
        ensure_atleast_bivariate(y=y)
        validate_response_support(self, y=y)
        _ = type_of_target(y, raise_unknown=True)

        # Prepare the estimator
        self._prepare_estimator()

        # Set fixed values
        self.dim_ = y.shape[1]
        self.n_observations_ = y.shape[0]
        self.n_training_ = calculate_effective_training_length(
            forget=self.learning_rate, n_obs=self.n_observations_
        )

        self.n_dist_elements_ = self.distribution.fitted_elements(self.dim_)
        # Validate the equation and set the method
        # Ensure that we have the same number of lambda_n for all methods per distr param.
        self._method = self._prepare_estimation_method(y=y)
        self._check_path_based_estimation_methods()
        self._equation = self.validate_equation(self.equation)
        self.n_features_ = self._get_number_of_covariates(X)

        # Without prior information, the optimal ADR regularization is
        # The largest model
        self.adr_steps_ = self.distribution.get_regularization_size(self.dim_)
        if self.max_regularisation_size is not None:
            self.adr_steps_ = np.fmin(self.adr_steps_, self.max_regularisation_size)

        self.optimal_adr_ = self.adr_steps_

        # Empty coefficient dictionaries
        self.coef_ = {}
        self.coef_path_ = {}

        # Handle scaling
        self._scaler.fit(X)
        X_scaled = self._scaler.transform(X=X)

        # Create the regularization mask
        # For the "lower estimation" i.e. the CD path.
        self.is_regularized_ = {
            p: {
                k: np.repeat(True, self.n_features_[p][k])
                for k in range(self.n_dist_elements_[p])
            }
            for p in range(self.distribution.n_params)
        }
        for p in range(self.distribution.n_params):
            for k in range(self.n_dist_elements_[p]):
                if not self._regularize_intercept[p]:
                    if self._fit_intercept[p]:
                        self.is_regularized_[p][k][0] = False

        # Iteration index for the elements of each distribution parameter
        self._iter_index = {
            p: self._make_iteration_indices(p)
            for p in range(self.distribution.n_params)
        }

        self._prepare_adr_regularization()
        theta = self._make_initial_theta(y)

        # Current information
        self._current_likelihood = np.array([
            np.sum(self.distribution.logpdf(y=y, theta=theta[a]))
            for a in range(self.adr_steps_)
        ])
        self._model_selection = {
            p: {a: {} for a in range(self.adr_steps_)}
            for p in range(self.distribution.n_params)
        }
        self._x_gram = {
            p: {
                k: np.empty((
                    self.adr_steps_,
                    self.n_features_[p][k],
                    self.n_features_[p][k],
                ))
                for k in range(self.n_dist_elements_[p])
            }
            for p in range(self.distribution.n_params)
        }
        self._y_gram = {
            p: {
                k: np.empty((self.adr_steps_, self.n_features_[p][k]))
                for k in range(self.n_dist_elements_[p])
            }
            for p in range(self.distribution.n_params)
        }
        self.coef_path_ = {
            p: {
                k: np.zeros((
                    self.adr_steps_,
                    self._lambda_n[p],
                    self.n_features_[p][k],
                ))
                for k in range(self.n_dist_elements_[p])
            }
            for p in range(self.distribution.n_params)
        }
        self.coef_ = {
            p: {
                k: np.zeros((self.adr_steps_, self.n_features_[p][k]))
                for k in range(self.n_dist_elements_[p])
            }
            for p in range(self.distribution.n_params)
        }

        # Some information about the different iterations
        self.iteration_count_ = np.zeros(
            (self.max_iterations_outer, self.distribution.n_params, self.adr_steps_),
            dtype=int,
        )
        self.iteration_likelihood_ = np.zeros((
            self.max_iterations_outer,
            self.max_iterations_inner,
            self.distribution.n_params,
            self.adr_steps_,
        ))
        # Call the fit
        self._outer_fit(X=X_scaled, y=y, theta=theta)
        self._print_message(message="Finished fitting distribution parameters.")

        return self

    def _outer_fit(self, X, y, theta):
        adr_start = time.time()

        for a in range(self.adr_steps_):
            adr_it_start = time.time()
            outer_start = time.time()

            global_old_likelihood = 0
            converged = False
            decreasing = False

            for outer_iteration in range(self.max_iterations_outer):
                outer_it_start = time.time()
                # Main outer loop
                for p in range(self.distribution.n_params):
                    message = (
                        f"Outer Iteration: {outer_iteration}, "
                        f"fitting Distribution parameter {p}, AD-r step {a}"
                    )
                    self._print_message(level=2, message=message)
                    theta = self._inner_fit(
                        X=X,
                        y=y,
                        theta=theta,
                        outer_iteration=outer_iteration,
                        p=p,
                        a=a,
                    )
                # Get the global likelihood
                global_likelihood = self._current_likelihood[a]

                converged, _ = self._check_outer_convergence(
                    global_old_likelihood, global_likelihood
                )

                # start value for the next AD-R step
                if (
                    converged
                    | decreasing
                    | (outer_iteration == self.max_iterations_outer - 1)
                ):
                    if a < (self.adr_steps_ - 1):
                        for p in range(self.distribution.n_params):
                            if not self.distribution._regularization_allowed[p]:
                                theta[(a + 1)][p] = copy.deepcopy(theta[a][p])
                            if self.distribution._regularization_allowed[p]:
                                theta[(a + 1)][p] = self._get_next_adr_start_values(
                                    theta, p, a + 1
                                )

                # Timings
                outer_it_end = time.time()
                outer_it_time = outer_it_end - outer_it_start
                outer_it_avg = (outer_it_end - outer_start) / (outer_iteration + 1)
                outer_pred_time = outer_it_time * (
                    self.max_iterations_outer - outer_iteration - 1
                )
                message = (
                    f"Last outer iteration {outer_iteration} took {round(outer_it_time, 1)} sec. "
                    f"Average outer iteration took {round(outer_it_avg, 1)} sec. "
                    f"Expected to be finished in max {round(outer_pred_time, 1)} sec. "
                )
                self._print_message(level=2, message=message)

                # End timing for the iteration
                adr_it_end = time.time()

                if converged:
                    break
                else:
                    global_old_likelihood = global_likelihood

            # Timings
            adr_it_last = adr_it_end - adr_it_start
            adr_it_avg = (adr_it_end - adr_start) / (a + 1)
            message = (
                f"Last ADR iteration {a} took {round(adr_it_last, 1)} sec. "
                f"Average ADR iteration took {round(adr_it_avg, 1)} sec. "
            )
            self._print_message(level=1, message=message)

            # Calculate the improvement
            if self.early_stopping_criteria == "ll":
                # Check the lilelihood for early stopping
                self.improvement_abs_ = -np.diff(self._current_likelihood)
                self.improvement_abs_scaled_ = -self.improvement_abs_ / self.n_training_
                self.improvement_rel_ = (
                    self.improvement_abs_ / self._current_likelihood[-1:]
                )

            elif self.early_stopping_criteria in ["aic", "bic", "hqc", "max"]:
                self._early_stopping_n_params = np.array([
                    self.count_nonzero_coef(self.coef_, a)
                    for a in range(self.adr_steps_)
                ])
                self.early_stopping_ic_ = InformationCriterion(
                    n_observations=self.n_training_,
                    n_parameters=self._early_stopping_n_params,
                    criterion=self.early_stopping_criteria,
                ).from_ll(self._current_likelihood)

                self.improvement_abs_ = -np.diff(self.early_stopping_ic_)
                self.improvement_abs_scaled_ = self.improvement_abs_
                self.improvement_rel_ = (
                    self.improvement_abs_ / self.early_stopping_ic_[-1:]
                )
            else:
                raise ValueError(
                    "Did not recognize criteria AD-R regularization stopping criteria."
                )

            if self.early_stopping and (a > 0) and (a < self.adr_steps_ - 1):
                # In the last step, it does not make sense to "early stop"

                if (
                    self.improvement_abs_scaled_[a - 1] < self.early_stopping_abs_tol
                ) or (self.improvement_rel_[a - 1] < self.early_stopping_rel_tol):
                    message = (
                        f"Early stopping due to AD-r-regression. "
                        f"Last increase in r lead to relative improvement: {self.improvement_rel_[a - 1]}, scaled absolute improvement {self.improvement_abs_scaled_[a - 1]}",
                    )
                    self._print_message(level=1, message=message)

                    if self.improvement_rel_[a - 1] > 0:
                        self.optimal_adr_ = a
                    elif self.improvement_rel_[a - 1] < 0:
                        self.optimal_adr_ = a - 1

                    self.improvement_abs_[(a + 1) :] = 0
                    self.improvement_rel_[(a + 1) :] = 0
                    self.improvement_abs_scaled_[(a + 1) :] = 0
                    break
            else:
                # The largest theta is the optimal one
                # But might be overfit
                self.optimal_adr_ = a

        self.theta_ = theta
        self.optimal_theta_ = self.theta_[self.optimal_adr_]
        self.last_fit_adr_max_ = self.optimal_adr_

    @staticmethod
    def count_nonzero_coef(beta, adr):
        non_zero = 0
        for p, coef in beta.items():
            non_zero += int(np.sum([np.sum(c[adr, :] != 0) for k, c in coef.items()]))
        return non_zero

    def _count_coef_to_be_fitted_param(self, adr: int, param: int, k: int):
        """Count the coefficients that should be fitted for this param"""
        idx_k = self._iter_index[param].tolist().index(k)
        if self.distribution._regularization_allowed[param]:
            count = 0
            for kk in self._iter_index[param][idx_k:]:
                count += int(not self._is_element_adr_regularized(param, kk, adr))
        else:
            count = len(self._iter_index[param][idx_k:])
        return count

    def count_coef_to_be_fitted(
        self, outer_iteration: int, inner_iteration: int, adr: int, param: int, k: int
    ):
        """Count all coefficients that should be fitted."""
        if (outer_iteration > 0) or (inner_iteration > 0):
            count = 0
        else:
            count = 0
            for p in range(param, self.distribution.n_params):
                count += self._count_coef_to_be_fitted_param(
                    adr=adr, param=p, k=k if p == param else 0
                )
        return count

    @staticmethod
    def _check_convergence(old_value, new_value, rel_tol, abs_tol) -> Tuple[bool, bool]:
        # Account for floating point acc
        if np.abs(new_value - old_value) / np.abs(new_value) < rel_tol:
            converged = True
            decreasing = False
        elif np.abs(new_value - old_value) < abs_tol:
            converged = True
            decreasing = False
        elif (new_value < old_value) & (np.abs(new_value - old_value) > 1e-10):
            converged = False
            decreasing = True
        else:
            converged = False
            decreasing = False
        return converged, decreasing

    def _check_inner_convergence(self, old_value, new_value) -> Tuple[bool, bool]:
        return self._check_convergence(
            old_value=old_value,
            new_value=new_value,
            rel_tol=self.rel_tol_inner,
            abs_tol=self.abs_tol_inner,
        )

    def _check_outer_convergence(self, old_value, new_value) -> Tuple[bool, bool]:
        return self._check_convergence(
            old_value=old_value,
            new_value=new_value,
            rel_tol=self.rel_tol_outer,
            abs_tol=self.abs_tol_outer,
        )

    def get_dampened_prediction(
        self,
        prediction: np.ndarray,
        eta: np.ndarray,
        outer_iteration: int,
        inner_iteration: int,
        param: int,
    ):
        if (outer_iteration == 0) & (inner_iteration < self._dampen_estimation[param]):
            out = (prediction * self._dampen_estimation[param] + eta) / (
                self._dampen_estimation[param] + 1
            )
        else:
            out = prediction
        return out

    def _inner_fit(self, y, X, theta, outer_iteration, a, p):
        converged = False
        decreasing = False
        old_likelihood = self._current_likelihood[a]

        weights_forget = init_forget_vector(
            forget=self.learning_rate,
            size=self.n_observations_,
        )

        for inner_iteration in range(self.max_iterations_inner):
            # If the likelihood is at some point decreasing, we're breaking
            # Hence we need to store previous iteration values:
            if (inner_iteration > 0) | (outer_iteration > 0):
                prev_theta = copy.copy(theta)
                prev_x_gram = copy.copy(self._x_gram[p])
                prev_y_gram = copy.copy(self._y_gram[p])
                prev_model_selection = copy.copy(self._model_selection)
                prev_beta = copy.copy(self.coef_)
                prev_beta_path = copy.copy(self.coef_path_)

            # This will check if we
            if (inner_iteration == 0) and (outer_iteration == 0) & (a == 0):
                theta[a] = self.distribution.set_initial_guess(y, theta[a], p)
                theta = self._handle_path_regularization(theta=theta, p=p, a=a)

            # Iterate through all elements of the distribution parameter
            for k in self._iter_index[p]:
                if self.debug:
                    print("Fitting", outer_iteration, inner_iteration, p, k, a)

                if self._is_element_adr_regularized(p=p, k=k, a=a):
                    self.coef_[p][k][a] = np.zeros(self.n_features_[p][k])
                    self.coef_path_[p][k][a] = np.zeros((
                        self._lambda_n[p],
                        self.n_features_[p][k],
                    ))
                else:
                    eta = self.distribution.link_function(theta[a][p], p)
                    eta = self.distribution.cube_to_flat(eta, param=p)

                    # Derivatives wrt to the parameter
                    dl1dp1 = self.distribution.element_dl1_dp1(
                        y, theta=theta[a], param=p, k=k
                    )
                    dl2dp2 = self.distribution.element_dl2_dp2(
                        y, theta=theta[a], param=p, k=k, clip=False
                    )

                    dl1_link = self.distribution.link_function_derivative(
                        theta[a][p], p
                    )
                    dl2_link = self.distribution.link_function_second_derivative(
                        theta[a][p],
                        p,
                    )
                    dl1_link = self.distribution.cube_to_flat(dl1_link, param=p)
                    dl1_link = dl1_link[:, k]
                    dl2_link = self.distribution.cube_to_flat(dl2_link, param=p)
                    dl2_link = dl2_link[:, k]

                    dl1_deta1 = dl1dp1 * (1 / dl1_link)
                    dl2_deta2 = (dl2dp2 * dl1_link - dl1dp1 * dl2_link) / dl1_link**3

                    wt = np.fmax(-dl2_deta2, 1e-10)
                    wv = eta[:, k] + dl1_deta1 / wt

                    # Create the more arrays
                    x = make_model_array(
                        X=X,
                        eq=self._equation[p][k],
                        fit_intercept=self._fit_intercept[p],
                    )
                    if self.debug:
                        print(
                            "Rank of X",
                            outer_iteration,
                            inner_iteration,
                            p,
                            k,
                            a,
                            np.linalg.matrix_rank(x),
                        )

                    self._x_gram[p][k][a] = self._method[p][k].init_x_gram(
                        X=x,
                        weights=wt ** self._weight_delta[p],
                        forget=self._forget[p],
                    )
                    self._y_gram[p][k][a] = (
                        self._method[p][k]
                        .init_y_gram(
                            X=x,
                            y=wv,
                            weights=wt ** self._weight_delta[p],
                            forget=self._forget[p],
                        )
                        .squeeze()
                    )

                    if self._method[p][k]._path_based_method:
                        self.coef_path_[p][k][a] = self._method[p][k].fit_beta_path(
                            x_gram=self._x_gram[p][k][a],
                            y_gram=self._y_gram[p][k][a][:, None],
                            is_regularized=self.is_regularized_[p][k],
                        )
                        eta_elem = x @ self.coef_path_[p][k][a].T
                        theta_elem = self.distribution.element_link_inverse(
                            eta_elem, param=p, k=k, d=self.dim_
                        )

                        opt_ic = self._fit_model_selection(
                            y=y,
                            theta_fit=theta_elem,
                            theta=theta,
                            outer_iteration=outer_iteration,
                            inner_iteration=inner_iteration,
                            a=a,
                            k=k,
                            param=p,
                        )
                        # select optimal beta and theta
                        self.coef_[p][k][a] = self.coef_path_[p][k][a][opt_ic, :]
                        theta[a] = self.distribution.set_theta_element(
                            theta[a], theta_elem[:, opt_ic], param=p, k=k
                        )
                    else:
                        self.coef_path_[p][k][a] = None
                        self.coef_[p][k][a] = self._method[p][k].fit_beta(
                            x_gram=self._x_gram[p][k][a],
                            y_gram=self._y_gram[p][k][a][:, None],
                            is_regularized=self.is_regularized_[p][k],
                        )

                    eta[:, k] = self.get_dampened_prediction(
                        prediction=np.squeeze(x @ self.coef_[p][k][a]),
                        eta=eta[:, k],
                        inner_iteration=inner_iteration,
                        outer_iteration=outer_iteration,
                        param=p,
                    )
                    theta[a][p] = self.distribution.link_inverse(
                        self.distribution.flat_to_cube(eta, param=p), param=p
                    )
                    if (self._overshoot_correction[p] is not None) and (
                        inner_iteration + outer_iteration < 1
                    ):
                        theta[a] = self.distribution.set_theta_element(
                            theta[a],
                            theta[a][p][:, k] + self.overshoot_correction[p][k],
                            param=p,
                            k=k,
                        )

                self._current_likelihood[a] = (
                    self.distribution.logpdf(y, theta=theta[a]) * weights_forget
                ).sum()

            ## Check the most important convergence measures here now
            if inner_iteration == (self.max_iterations_inner - 1):
                warnings.warn(
                    "Reached max inner iterations. Algorithm may or may not be converged."
                )
            self.iteration_count_[outer_iteration, p, a] = inner_iteration
            self.iteration_likelihood_[outer_iteration, inner_iteration, p, a] = (
                self._current_likelihood[a]
            )
            message = (
                f"Outer iteration: {outer_iteration}, inner iteration {inner_iteration}, parameter {p}, AD-R {a}:"
                f"current likelihood: {self._current_likelihood[a]},"
                f"previous iteration likelihood {self.iteration_likelihood_[outer_iteration, inner_iteration - 1, p, a] if inner_iteration > 0 else self._current_likelihood[a]}"
            )
            self._print_message(level=2, message=message)

            if (inner_iteration > 0) & (
                (inner_iteration > self._dampen_estimation[p]) | (outer_iteration > 0)
            ):
                converged, decreasing = self._check_inner_convergence(
                    old_value=old_likelihood, new_value=self._current_likelihood[a]
                )

                if converged:
                    break
                else:
                    # For the next iteration
                    old_likelihood = self._current_likelihood[a]

            # If the LL is decreasing, we're resetting to the previous iteration
            if ((outer_iteration > 0) | (inner_iteration > 1)) & decreasing:
                warnings.warn("Likelihood is decreasing. Breaking.")
                # Reset to values from the previous iteration
                theta = prev_theta
                self._model_selection = prev_model_selection
                self._x_gram[p] = prev_x_gram
                self._y_gram[p] = prev_y_gram
                self.coef_ = prev_beta
                self.coef_path_ = prev_beta_path
                self._current_likelihood[a] = old_likelihood
                break

        return theta

    def _fit_model_selection(
        self,
        y,
        theta_fit,
        theta,
        outer_iteration,
        inner_iteration,
        a: int,
        k: int,
        param: int,
    ):
        weights_forget = init_forget_vector(
            self.learning_rate,
            self.n_observations_,
        )
        # Model selection
        if self.ic == "max":
            opt_ic = self._lambda_n[param] - 1
        else:
            theta_ll = copy.deepcopy(theta[a])
            theta_elem_delta = np.diff(theta_fit, axis=1)
            theta_ms = self.distribution.set_theta_element(
                theta_ll, theta_fit[:, 0], param=param, k=k
            )
            approx_ll = np.sum(self.distribution.logpdf(y, theta_ll) * weights_forget)
            approx_ll = np.repeat(approx_ll, 100)
            for l_idx in range(1, self._lambda_n[param]):
                theta_ms = self.distribution.set_theta_element(
                    theta_ll, theta_fit[:, l_idx], param=param, k=k
                )
                if self.approx_fast_model_selection:
                    approx_ll[l_idx] = approx_ll[l_idx - 1] + np.sum(
                        self.distribution.element_dl1_dp1(y, theta_ms, param=param, k=k)
                        * theta_elem_delta[:, l_idx - 1]
                        * weights_forget
                    )
                else:
                    approx_ll[l_idx] = np.sum(
                        self.distribution.logpdf(
                            y,
                            theta=theta_ms,
                        )
                        * weights_forget
                    )

            # Count number of nonzero coefficients
            # Subtract current beta if already fitted
            # If in the first iteration, add intercept
            # for all to-be-fitted parameters
            # that are not AD-R regularized
            nonzero = self.count_nonzero_coef(self.coef_, adr=a)
            nonzero = nonzero + int(np.sum(self.coef_[param][k][a, :] != 0))
            nonzero = nonzero + self.count_coef_to_be_fitted(
                outer_iteration, inner_iteration, adr=a, param=param, k=k
            )
            ic = InformationCriterion(
                n_observations=self.n_observations_,
                n_parameters=nonzero,
                criterion=self.ic,
            ).from_ll(log_likelihood=approx_ll)
            self._model_selection[param][a][k] = {
                "ll": approx_ll,
                "non_zero": nonzero,
                "ic": ic,
            }
            opt_ic = np.argmin(ic)

        return opt_ic

    def _update_model_selection(
        self,
        y,
        theta_fit: np.ndarray,
        theta: Dict,
        outer_iteration: int,
        inner_iteration: int,
        a: int,
        k: int,
        param: int,
    ) -> int:
        """
        Update the model selection.

        Returns the optimal IC's index.
        """
        weights_forget = init_forget_vector(
            self.learning_rate,
            self.n_observations_step_,
        )

        if self.ic == "max":
            opt_ic = self._lambda_n[param] - 1
        else:
            # Model selection
            theta_ll = copy.deepcopy(theta[a])

            theta_elem_delta = np.diff(theta_fit, axis=1)
            theta_ms = self.distribution.set_theta_element(
                theta_ll, theta_fit[:, 0], param=param, k=k
            )
            approx_ll = np.sum(self.distribution.logpdf(y, theta_ll) * weights_forget)
            approx_ll = np.repeat(approx_ll, 100)
            for l_idx in range(1, self._lambda_n[param]):
                theta_ms = self.distribution.set_theta_element(
                    theta_ll, theta_fit[:, l_idx], param=param, k=k
                )
                if self.approx_fast_model_selection:
                    approx_ll[l_idx] = approx_ll[l_idx - 1] + np.sum(
                        (
                            self.distribution.element_dl1_dp1(
                                y, theta_ms, param=param, k=k
                            )
                            * theta_elem_delta[:, l_idx - 1]
                        )
                        * weights_forget
                    )
                else:
                    approx_ll[l_idx] = np.sum(
                        self.distribution.logpdf(
                            y,
                            theta=theta_ms,
                        )
                        * weights_forget
                    )
            approx_ll = approx_ll + (
                self._model_selection_old[param][a][k]["ll"]
                * (1 - self.learning_rate) ** self.n_observations_step_
            )
            # Count number of nonzero coefficients
            # Subtract current beta if already fitted
            # If in the first iteration, add intercept
            # for all to-be-fitted parameters
            # that are not AD-R regularized
            nonzero = self.count_nonzero_coef(self.coef_, adr=a)
            nonzero = nonzero + int(np.sum(self.coef_[param][k][a, :] != 0))
            nonzero = nonzero + self.count_coef_to_be_fitted(
                outer_iteration, inner_iteration, adr=a, param=param, k=k
            )
            ic = InformationCriterion(
                n_observations=self.n_observations_,
                n_parameters=nonzero,
                criterion=self.ic,
            ).from_ll(log_likelihood=approx_ll)
            opt_ic = np.argmin(ic)
            self._model_selection[param][a][k] = {
                "ll": approx_ll,
                "non_zero": nonzero,
                "ic": ic,
            }

        return opt_ic

    # Different UV - MV
    def predict(
        self,
        X: Optional[np.ndarray] = None,
    ) -> Dict[int, np.ndarray]:
        """
        Predict the location parameter.

        Parameters
        ----------
        X : np.ndarray, optional
            The input data. If None, will use a default value of ones.

        """

        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        if X is None:
            X_scaled = np.ones((1, 1))
            N = 1
            self._print_message("X is None. Prediction will have length 1.")
        else:
            X_scaled = self._scaler.transform(X=X)
            N = X.shape[0]

        array = np.zeros((N, self.n_dist_elements_[0]))
        for k in range(self.n_dist_elements_[0]):
            array[:, k] = (
                make_model_array(
                    X=X_scaled,
                    eq=self._equation[0][k],
                    fit_intercept=self._fit_intercept[0],
                )
                @ self.coef_[0][k][self.optimal_adr_, :]
            ).squeeze()
        out = self.distribution.flat_to_cube(array, 0)
        out = self.distribution.link_inverse(out, 0)
        return out

    # Different UV - MV
    def predict_distribution_parameters(
        self,
        X: Optional[np.ndarray] = None,
    ) -> Dict[int, np.ndarray]:
        """Predict the distribution parameters.

        Args:
            X (Optional[np.ndarray], optional): Covariate or feature matrix. Defaults to None.

        Returns:
            Dict[int, np.ndarray]: Return the predicted distribution parameters as a dictionary.
        """

        check_is_fitted(self)
        X = validate_data(
            self,
            X=X,
            reset=False,
            dtype=[np.float64, np.float32],
        )

        if X is None:
            X_scaled = np.ones((1, 1))
            N = 1
            self._print_message("X is None. Prediction will have length 1.")
        else:
            X_scaled = self._scaler.transform(X=X)
            N = X.shape[0]
        out = {}

        for p in range(self.distribution.n_params):
            array = np.zeros((N, self.n_dist_elements_[p]))
            for k in range(self.n_dist_elements_[p]):
                array[:, k] = (
                    make_model_array(
                        X=X_scaled,
                        eq=self._equation[p][k],
                        fit_intercept=self._fit_intercept[p],
                    )
                    @ self.coef_[p][k][self.optimal_adr_, :]
                ).squeeze()
            out[p] = self.distribution.flat_to_cube(array, p)
            out[p] = self.distribution.link_inverse(out[p], p)
        return out

    def predict_all_adr(
        self,
        X: Optional[np.ndarray] = None,
    ) -> Dict[int, np.ndarray]:
        """Predict all regualarized fits in one go.

        Args:
            X (Optional[np.ndarray], optional): Covariate or feature matrix. Defaults to None.

        Returns:
            Dict[int, np.ndarray]: Return the predicted distribution parameters for all AD-R steps as a dictionary. The structure is
            {adr_step: {param: np.ndarray}} where `adr_step` is the AD-R step index and `param` is the distribution parameter index.
        """

        check_is_fitted(self)
        X = validate_data(self, X=X, reset=False, dtype=[np.float64, np.float32])

        if X is None:
            X_scaled = np.ones((1, 1))
            N = 1
            self._print_message("X is None. Prediction will have length 1.")
        else:
            X_scaled = self._scaler.transform(X=X)
            N = X.shape[0]
        out = {}
        for a in range(self.adr_steps_):
            out[a] = {}
            if a <= self.optimal_adr_:
                for p in range(self.distribution.n_params):
                    array = np.zeros((N, self.n_dist_elements_[p]))
                    for k in range(self.n_dist_elements_[p]):
                        array[:, k] = (
                            make_model_array(
                                X=X_scaled,
                                eq=self._equation[p][k],
                                fit_intercept=self._fit_intercept[p],
                            )
                            @ self.coef_[p][k][a, :]
                        ).squeeze()
                    out[a][p] = self.distribution.flat_to_cube(array, p)
                    out[a][p] = self.distribution.link_inverse(out[a][p], p)
            else:
                out[a] = copy.deepcopy(out[self.optimal_adr_])

        return out

    def partial_fit(self, X: np.ndarray, y: np.ndarray):
        """
        Align ondil with the scikit-learn API for partial fitting.

        The first partial fit will call `fit`, and subsequent calls will call `update`.
        Allows furthermore to use the sklearn testing framework.

        Overwrites the base class method to avoid sample_weights. This estimator does not support sample weights.

        Parameters
        ----------
        X : np.ndarray
            The input data.
        y : np.ndarray
            The target values.
        Returns
        -------
        self : Estimator
            The fitted estimator.

        """
        if self.is_fitted:
            self.update(X=X, y=y)
        else:
            self.fit(X=X, y=y)
        return self

    # Different UV - MV
    @_fit_context(prefer_skip_nested_validation=True)
    def update(
        self, X: np.ndarray, y: np.ndarray
    ) -> "MultivariateOnlineDistributionalRegressionPath":
        """
        Updates the estimator with new observations.

        This method validates the input data, updates internal counters for the number of observations,
        recalculates the effective training length based on the learning rate, and updates the model's
        likelihood. It also scales the input features, stores previous states for model selection and
        likelihood, and performs an outer update step with the scaled data.

        Parameters
        ----------
        X : np.ndarray
            Input feature matrix of shape (n_samples, n_features).
        y : np.ndarray
            Target values of shape (n_samples, n_outputs).
        Returns
        -------
        self : object
            Returns the updated estimator instance.
        """

        X, y = validate_data(
            self,
            X=X,
            y=y,
            reset=False,
            dtype=[np.float64, np.float32],
            multi_output=True,
        )
        ensure_atleast_bivariate(y=y)
        validate_response_support(self, y=y)

        self.n_observations_ += y.shape[0]
        self.n_observations_step_ = y.shape[0]
        self.n_training_ = calculate_effective_training_length(
            forget=self.learning_rate, n_obs=self.n_observations_
        )
        theta = self.predict_all_adr(X)
        self._scaler.update(X=X)
        X_scaled = self._scaler.transform(X=X)

        self._x_gram_old = copy.deepcopy(self._x_gram)
        self._y_gram_old = copy.deepcopy(self._y_gram)
        self._model_selection_old = copy.deepcopy(self._model_selection)
        self._old_likelihood = self._current_likelihood + 0
        self._old_likelihood_discounted = (
            1 - self.learning_rate
        ) ** self.n_observations_step_ * self._old_likelihood
        self._current_likelihood = self._old_likelihood_discounted + np.array([
            np.sum(
                self.distribution.logpdf(y=y, theta=theta[a])
                * init_forget_vector(self.learning_rate, y.shape[0])
            )
            for a in range(self.adr_steps_)
        ])
        self._outer_update(X=X_scaled, y=y, theta=theta)

        return self

    # Different UV - MV
    def _inner_update(self, X, y, theta, outer_iteration, a, p):
        converged = False
        decreasing = False
        old_likelihood = self._current_likelihood[a]
        weights_forget = init_forget_vector(
            self.learning_rate, self.n_observations_step_
        )

        for inner_iteration in range(self.max_iterations_inner):
            # If the likelihood is at some point decreasing, we're breaking
            # Hence we need to store previous iteration values:
            if (inner_iteration > 0) | (outer_iteration > 0):
                prev_theta = copy.copy(theta)
                prev_x_gram = copy.copy(self._x_gram[p])
                prev_y_gram = copy.copy(self._y_gram[p])
                prev_model_selection = copy.copy(self._model_selection)
                prev_beta = copy.copy(self.coef_)
                prev_beta_path = copy.copy(self.coef_path_)

            for k in self._iter_index[p]:
                # Handle AD-R Regularization
                if self._is_element_adr_regularized(p=p, k=k, a=a):
                    self.coef_[p][k][a] = np.zeros(self.n_features_[p][k])
                    self.coef_path_[p][k][a] = np.zeros((
                        self._lambda_n[p],
                        self.n_features_[p][k],
                    ))

                else:
                    eta = self.distribution.link_function(theta[a][p], p)
                    eta = self.distribution.cube_to_flat(eta, param=p)

                    # Derivatives wrt to the parameter
                    dl1dp1 = self.distribution.element_dl1_dp1(
                        y, theta=theta[a], param=p, k=k
                    )
                    dl2dp2 = self.distribution.element_dl2_dp2(
                        y, theta=theta[a], param=p, k=k
                    )

                    dl1_link = self.distribution.link_function_derivative(
                        theta[a][p], p
                    )
                    dl2_link = self.distribution.link_function_second_derivative(
                        theta[a][p], p
                    )
                    dl1_link = self.distribution.cube_to_flat(dl1_link, param=p)
                    dl1_link = dl1_link[:, k]
                    dl2_link = self.distribution.cube_to_flat(dl2_link, param=p)
                    dl2_link = dl2_link[:, k]

                    dl1_deta1 = dl1dp1 * (1 / dl1_link)
                    dl2_deta2 = (dl2dp2 * dl1_link - dl1dp1 * dl2_link) / dl1_link**3

                    wt = np.fmax(-dl2_deta2, 1e-10)
                    wv = eta[:, k] + dl1_deta1 / wt

                    # Make model arrays
                    x = make_model_array(
                        X=X,
                        eq=self._equation[p][k],
                        fit_intercept=self._fit_intercept[p],
                    )
                    self._x_gram[p][k][a] = self._method[p][k].update_x_gram(
                        gram=self._x_gram_old[p][k][a],
                        X=x,
                        weights=wt ** self._weight_delta[p],
                        forget=self._forget[p],
                    )
                    self._y_gram[p][k][a] = (
                        self._method[p][k]
                        .update_y_gram(
                            gram=np.expand_dims(self._y_gram_old[p][k][a], -1),
                            X=x,
                            y=wv,
                            weights=wt ** self._weight_delta[p],
                            forget=self._forget[p],
                        )
                        .squeeze()
                    )
                    if self._method[p][k]._path_based_method:
                        self.coef_path_[p][k][a] = self._method[p][k].update_beta_path(
                            x_gram=self._x_gram[p][k][a],
                            y_gram=self._y_gram[p][k][a][:, None],
                            beta_path=self.coef_path_[p][k][a],
                            is_regularized=self.is_regularized_[p][k],
                        )
                        eta_elem = x @ self.coef_path_[p][k][a].T
                        theta_elem = self.distribution.element_link_inverse(
                            eta_elem, param=p, k=k, d=self.dim_
                        )

                        opt_ic = self._update_model_selection(
                            y=y,
                            theta=theta,
                            theta_fit=theta_elem,
                            outer_iteration=outer_iteration,
                            inner_iteration=inner_iteration,
                            param=p,
                            k=k,
                            a=a,
                        )
                        # Select the optimal beta
                        self.coef_[p][k][a] = self.coef_path_[p][k][a][opt_ic, :]
                        theta[a] = self.distribution.set_theta_element(
                            theta[a], theta_elem[:, opt_ic], param=p, k=k
                        )
                    else:
                        self.coef_path_[p][k][a] = None
                        self.coef_[p][k][a] = self._method[p][k].update_beta(
                            x_gram=self._x_gram[p][k][a],
                            y_gram=self._y_gram[p][k][a][:, None],
                            beta=self.coef_[p][k][a],
                            is_regularized=self.is_regularized_[p][k],
                        )
                        self.coef_[p][k][a] = (
                            self._x_gram[p][k][a] @ self._y_gram[p][k][a]
                        )

                    # Calculate the other stuff
                    eta[:, k] = np.squeeze(x @ self.coef_[p][k][a])
                    theta[a][p] = self.distribution.link_inverse(
                        self.distribution.flat_to_cube(eta, param=p), param=p
                    )

                self._current_likelihood[a] = (
                    np.sum(self.distribution.logpdf(y, theta=theta[a]) * weights_forget)
                    + self._old_likelihood_discounted[a]
                )

            self.iteration_count_[outer_iteration, p] = inner_iteration
            self.iteration_likelihood_[outer_iteration, inner_iteration, p, a] = (
                self._current_likelihood[a]
            )

            # Are we in the last iteration
            if inner_iteration == (self.max_iterations_inner - 1):
                warnings.warn(
                    "Reached max inner iterations. Algorithm may or may not be converged."
                )

            # Are we converged
            if inner_iteration > 0:
                converged, decreasing = self._check_inner_convergence(
                    old_value=old_likelihood, new_value=self._current_likelihood[a]
                )

                if converged:
                    break
                else:
                    # For the next iteration
                    old_likelihood = self._current_likelihood[a]

            # Are we diverging?
            if ((outer_iteration > 0) | (inner_iteration > 1)) & decreasing:
                warnings.warn("Likelihood is decreasing. Breaking.")
                # Reset to values from the previous iteration
                theta = prev_theta
                self._model_selection = prev_model_selection
                self._x_gram[p] = prev_x_gram
                self._y_gram[p] = prev_y_gram
                self.coef_ = prev_beta
                self.coef_path_ = prev_beta_path
                self._current_likelihood[a] = old_likelihood
                break

        return theta

    # Different UV - MV
    def _outer_update(self, X, y, theta):
        adr_start = time.time()
        for a in range(min(self.adr_steps_, self.last_fit_adr_max_ + 1)):
            adr_it_start = time.time()
            outer_start = time.time()

            global_old_likelihood = 0
            converged = False

            for outer_iteration in range(self.max_iterations_outer):
                outer_it_start = time.time()
                # Main Loop
                for p in range(self.distribution.n_params):
                    message = (
                        f"Outer Iteration: {outer_iteration}, "
                        f"fitting Distribution parameter {p}, AD-r step {a}"
                    )
                    self._print_message(level=2, message=message)
                    theta = self._inner_update(
                        X=X, y=y, theta=theta, outer_iteration=outer_iteration, p=p, a=a
                    )

                # Get global LL
                global_likelihood = self._current_likelihood[a]
                converged, _ = self._check_outer_convergence(
                    global_old_likelihood, global_likelihood
                )

                # start value for the next AD-R step
                if converged | (outer_iteration == self.max_iterations_outer - 1):
                    if a < (self.adr_steps_ - 1):
                        for p in range(self.distribution.n_params):
                            if not self.distribution._regularization_allowed[p]:
                                theta[(a + 1)][p] = copy.deepcopy(theta[a][p])
                            if self.distribution._regularization_allowed[p]:
                                theta[(a + 1)][p] = self._get_next_adr_start_values(
                                    theta, p, a + 1
                                )

                # Timings
                outer_it_end = time.time()
                outer_it_time = outer_it_end - outer_it_start
                outer_it_avg = (outer_it_end - outer_start) / (outer_iteration + 1)
                outer_pred_time = outer_it_time * (
                    self.max_iterations_outer - outer_iteration - 1
                )
                message = (
                    f"Last outer iteration {outer_iteration} took {round(outer_it_time, 1)} sec. "
                    f"Average outer iteration took {round(outer_it_avg, 1)} sec. "
                    f"Expected to be finished in max {round(outer_pred_time, 1)} sec. "
                )
                self._print_message(level=2, message=message)

                # End timing for the iteration
                adr_it_end = time.time()

                if converged:
                    break
                else:
                    global_old_likelihood = global_likelihood

            # Next AD-R iteration
            adr_it_last = adr_it_end - adr_it_start
            adr_it_avg = (adr_it_end - adr_start) / (a + 1)

            message = (
                f"Last ADR iteration {a} took {round(adr_it_last, 1)} sec. "
                f"Average ADR iteration took {round(adr_it_avg, 1)} sec. ",
            )
            self._print_message(level=1, message=message)

            # Calculate the improvement
            if self.early_stopping_criteria == "ll":
                # Check the likelihood for early stopping
                self.improvement_abs_ = -np.diff(self._current_likelihood)
                self.improvement_abs_scaled_ = -self.improvement_abs_ / self.n_training_
                self.improvement_rel_ = (
                    self.improvement_abs_ / self._current_likelihood[-1:]
                )

            elif self.early_stopping_criteria in ["aic", "bic", "hqc", "max"]:
                self._early_stopping_n_params = np.array([
                    self.count_nonzero_coef(self.coef_, a)
                    for a in range(self.adr_steps_)
                ])
                self.early_stopping_ic_ = InformationCriterion(
                    n_observations=self.n_training_,
                    n_parameters=self._early_stopping_n_params,
                    criterion=self.early_stopping_criteria,
                ).from_ll(self._current_likelihood)

                self.improvement_abs_ = -np.diff(self.early_stopping_ic_)
                self.improvement_abs_scaled_ = self.improvement_abs_
                self.improvement_rel_ = (
                    self.improvement_abs_ / self.early_stopping_ic_[-1:]
                )
            else:
                raise ValueError(
                    "Did not recognize criteria AD-R regularization stopping criteria."
                )

            if (
                self.early_stopping
                and (a > 0)
                and (
                    a < min(self.adr_steps_ - 1, self.last_fit_adr_max_)
                )  # In the last step, it does not make sense to "early stop"
            ):
                if (
                    self.improvement_abs_scaled_[a - 1] < self.early_stopping_abs_tol
                ) or (self.improvement_rel_[a - 1] < self.early_stopping_rel_tol):
                    message = (
                        f"Early stopping due to AD-r-regression. "
                        f"Last increase in r lead to relative improvement: {self.improvement_rel_[a - 1]}, scaled absolute improvement {self.improvement_abs_scaled_[a - 1]}",
                    )

                    self._print_message(level=1, message=message)
                    if self.improvement_rel_[a - 1] > 0:
                        self.optimal_adr_ = a
                    elif self.improvement_rel_[a - 1] < 0:
                        self.optimal_adr_ = a - 1

                    self.improvement_abs_[(a + 1) :] = 0
                    self.improvement_rel_[(a + 1) :] = 0
                    self.improvement_abs_scaled_[(a + 1) :] = 0
                    break
            else:
                # The largest theta is the optimal one
                # But might be overfit
                self.optimal_adr_ = a

        self.theta_ = theta
        self.optimal_theta_ = self.theta_[self.optimal_adr_]
        self.last_fit_adr_max_ = self.optimal_adr_



class OutOfSupportError(ValueError):
    """Exception raised for values that are out of support."""

    def __init__(self, message="This value is out of support."):
        self.message = message
        super().__init__(self.message)


def check_matplotlib(has_mpl: bool) -> None:
    """Check if matplotlib is available."""
    if not has_mpl:
        raise ImportError(
            "Matplotlib is required for plotting. "
            "Please install it with 'pip install matplotlib'."
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Tuple

import numpy as np
import scipy.stats as stats
from sklearn.base import BaseEstimator

from . import HAS_MPL
from .base import DiagnosticDisplay
from .error import check_matplotlib
from .robust_math import SMALL_NUMBER

if TYPE_CHECKING:
    import matplotlib.pyplot as plt


class PITHistogramDisplay(DiagnosticDisplay):
    def __init__(self, X, y, unif):
        self.X_ = X
        self.y_ = y
        self.unif_ = unif

    @classmethod
    def from_predictions(
        cls,
        y,
        theta,
        distribution,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "PITHistogramDisplay":
        unif = distribution.cdf(y, theta)
        return cls(None, y, unif).plot(ax=ax, figsize=figsize, **kwargs)

    @classmethod
    def from_estimator(
        cls,
        estimator: BaseEstimator,
        X: np.ndarray,
        y: np.ndarray,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "PITHistogramDisplay":
        predictions = estimator.predict_distribution_parameters(X)
        unif = estimator.distribution.cdf(y, predictions)
        return cls(X, y, unif).plot(ax=ax, figsize=figsize, **kwargs)

    def plot(
        self,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "PITHistogramDisplay":
        check_matplotlib(HAS_MPL)
        import matplotlib.pyplot as plt  # noqa: F401

        bins = kwargs.pop(
            "bins", np.linspace(0, 1, min(round(np.sqrt(self.y_.shape[0])) + 1, 50))
        )
        density = kwargs.pop("density", True)
        color = kwargs.pop("color", "grey")
        edgecolor = kwargs.pop("edgecolor", "black")
        lw = kwargs.pop("lw", 0.5)

        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        ax.set_title("PIT Histogram")
        ax.hist(
            self.unif_,
            bins=bins,
            density=density,
            color=color,
            edgecolor=edgecolor,
            lw=lw,
        )
        ax.set_xlabel("Uniform Space")
        ax.set_ylabel("Density")
        ax.set_xlim(0, 1)
        ax.axhline(1, color="red")
        ax.grid()

        self.ax_ = ax
        self.figure_ = ax.figure

        return self


class QQDisplay(DiagnosticDisplay):
    def __init__(self, X, y, theoretical, empirical):
        self.X_ = X
        self.y_ = y
        self.theoretical_ = theoretical
        self.empirical_ = empirical

    @classmethod
    def from_predictions(
        cls,
        y,
        theta,
        distribution,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "QQDisplay":
        quantiles = distribution.cdf(y=y, theta=theta)
        quantiles = np.clip(quantiles, SMALL_NUMBER, 1 - SMALL_NUMBER)
        n = len(y)
        theoretical = np.linspace(1 / (n + 1), n / (n + 1), n)
        empirical = np.sort(quantiles)
        return QQDisplay(None, y, theoretical, empirical).plot(
            ax=ax, figsize=figsize, **kwargs
        )

    @classmethod
    def from_estimator(
        cls,
        estimator: BaseEstimator,
        X: np.ndarray,
        y: np.ndarray,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "QQDisplay":
        pred = estimator.predict_distribution_parameters(X)
        return cls.from_predictions(
            y, pred, estimator.distribution, ax, figsize, **kwargs
        )

    def plot(
        self,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "QQDisplay":
        check_matplotlib(HAS_MPL)
        import matplotlib.pyplot as plt  # noqa: F401

        color = kwargs.pop("color", "blue")
        s = kwargs.pop("s", 20)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        ax.scatter(self.theoretical_, self.empirical_, color=color, s=s, **kwargs)
        ax.plot([0, 1], [0, 1], color="red", lw=1)
        ax.set_xlabel("Theoretical Quantiles")
        ax.set_ylabel("Empirical Quantiles")
        ax.set_title("QQ Plot")
        ax.grid()
        self.ax_ = ax
        self.figure_ = ax.figure
        return self


class WormPlotDisplay(DiagnosticDisplay):
    def __init__(self, X, y, xx, yy, z, lower_bound, upper_bound):
        self.X_ = X
        self.y_ = y
        self.xx_ = xx
        self.yy_ = yy
        self.z_ = z
        self.lower_bound_ = lower_bound
        self.upper_bound_ = upper_bound

    @classmethod
    def from_predictions(
        cls,
        y,
        theta,
        distribution,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        level: float = 0.95,
        **kwargs,
    ) -> "WormPlotDisplay":
        quantiles = distribution.cdf(y=y, theta=theta)
        quantiles = np.clip(quantiles, SMALL_NUMBER, 1 - SMALL_NUMBER)
        residuals = stats.norm.ppf(quantiles)

        xx, yy = stats.probplot(residuals, fit=False)
        yy = yy - xx
        n = len(xx)
        z = np.linspace(np.min(xx), np.max(xx), n)
        p = stats.norm(loc=0, scale=1).cdf(z)
        se = (1 / stats.norm().pdf(z)) * (np.sqrt(p * (1 - p) / n))
        lower_bound = se * stats.norm.ppf((1 - level) / 2)
        upper_bound = se * stats.norm.ppf((1 + level) / 2)

        return cls(
            X=None,
            y=y,
            xx=xx,
            yy=yy,
            z=z,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
        ).plot(ax=ax, figsize=figsize, **kwargs)

    @classmethod
    def from_estimator(
        cls,
        estimator: BaseEstimator,
        X: np.ndarray,
        y: np.ndarray,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        level: float = 0.95,
        **kwargs,
    ) -> "WormPlotDisplay":
        pred = estimator.predict_distribution_parameters(X)
        return cls.from_predictions(
            y, pred, estimator.distribution, ax, figsize, level, **kwargs
        )

    def plot(
        self,
        ax: plt.Axes = None,
        figsize: Tuple[float, float] = (10, 5),
        **kwargs,
    ) -> "WormPlotDisplay":
        check_matplotlib(HAS_MPL)
        import matplotlib.pyplot as plt  # noqa: F401

        color = kwargs.pop("color", "blue")
        alpha = kwargs.pop("alpha", 0.2)
        if ax is None:
            _, ax = plt.subplots(figsize=figsize)
        ax.scatter(self.xx_, self.yy_, color=color, **kwargs)
        ax.plot(self.z_, self.lower_bound_, color="red", label="Lower confidence bound")
        ax.plot(self.z_, self.upper_bound_, color="red", label="Upper confidence bound")
        ax.fill_between(
            self.z_, self.lower_bound_, self.upper_bound_, color="grey", alpha=alpha
        )
        ax.axhline(0, color="black", lw=1, ls="--")
        ax.set_xlabel("Theoretical Quantiles")
        ax.set_ylabel("Empirical - Theoretical Quantiles")
        ax.set_title("Worm Plot (De-trended QQ Plot)")
        ax.grid()
        self.ax_ = ax
        self.figure_ = ax.figure
        return self


__ALL__ = ["PITHistogramDisplay", "QQDisplay", "WormPlotDisplay"]
